# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

import json

import requests
import responses
from nose.tools import (
  assert_equal,
  assert_is_instance,
  assert_is_none,
  assert_is_not_none,
  assert_not_equal,
  assert_raises
)

from gocardless_pro.errors import MalformedResponseError
from gocardless_pro import resources
from gocardless_pro import list_response

from .. import helpers
  

@responses.activate
def test_instalment_schedules_create_with_dates():
    fixture = helpers.load_fixture('instalment_schedules')['create_with_dates']
    helpers.stub_response(fixture)
    response = helpers.client.instalment_schedules.create_with_dates(*fixture['url_params'])
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.currency, body.get('currency'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.name, body.get('name'))
    assert_equal(response.payment_errors, body.get('payment_errors'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.total_amount, body.get('total_amount'))
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.mandate,
                 body.get('links')['mandate'])
    assert_equal(response.links.payments,
                 body.get('links')['payments'])

@responses.activate
def test_instalment_schedules_create_with_dates_new_idempotency_key_for_each_call():
    fixture = helpers.load_fixture('instalment_schedules')['create_with_dates']
    helpers.stub_response(fixture)
    helpers.client.instalment_schedules.create_with_dates(*fixture['url_params'])
    helpers.client.instalment_schedules.create_with_dates(*fixture['url_params'])
    assert_not_equal(responses.calls[0].request.headers.get('Idempotency-Key'),
                     responses.calls[1].request.headers.get('Idempotency-Key'))

def test_timeout_instalment_schedules_create_with_dates_idempotency_conflict():
    create_fixture = helpers.load_fixture('instalment_schedules')['create_with_dates']
    get_fixture = helpers.load_fixture('instalment_schedules')['get']
    with helpers.stub_timeout_then_idempotency_conflict(create_fixture, get_fixture) as rsps:
      response = helpers.client.instalment_schedules.create_with_dates(*create_fixture['url_params'])
      assert_equal(2, len(rsps.calls))

    assert_is_instance(response, resources.InstalmentSchedule)

@responses.activate
def test_timeout_instalment_schedules_create_with_dates_retries():
    fixture = helpers.load_fixture('instalment_schedules')['create_with_dates']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.instalment_schedules.create_with_dates(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)

def test_502_instalment_schedules_create_with_dates_retries():
    fixture = helpers.load_fixture('instalment_schedules')['create_with_dates']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.instalment_schedules.create_with_dates(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)
  

@responses.activate
def test_instalment_schedules_create_with_schedule():
    fixture = helpers.load_fixture('instalment_schedules')['create_with_schedule']
    helpers.stub_response(fixture)
    response = helpers.client.instalment_schedules.create_with_schedule(*fixture['url_params'])
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.currency, body.get('currency'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.name, body.get('name'))
    assert_equal(response.payment_errors, body.get('payment_errors'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.total_amount, body.get('total_amount'))
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.mandate,
                 body.get('links')['mandate'])
    assert_equal(response.links.payments,
                 body.get('links')['payments'])

@responses.activate
def test_instalment_schedules_create_with_schedule_new_idempotency_key_for_each_call():
    fixture = helpers.load_fixture('instalment_schedules')['create_with_schedule']
    helpers.stub_response(fixture)
    helpers.client.instalment_schedules.create_with_schedule(*fixture['url_params'])
    helpers.client.instalment_schedules.create_with_schedule(*fixture['url_params'])
    assert_not_equal(responses.calls[0].request.headers.get('Idempotency-Key'),
                     responses.calls[1].request.headers.get('Idempotency-Key'))

def test_timeout_instalment_schedules_create_with_schedule_idempotency_conflict():
    create_fixture = helpers.load_fixture('instalment_schedules')['create_with_schedule']
    get_fixture = helpers.load_fixture('instalment_schedules')['get']
    with helpers.stub_timeout_then_idempotency_conflict(create_fixture, get_fixture) as rsps:
      response = helpers.client.instalment_schedules.create_with_schedule(*create_fixture['url_params'])
      assert_equal(2, len(rsps.calls))

    assert_is_instance(response, resources.InstalmentSchedule)

@responses.activate
def test_timeout_instalment_schedules_create_with_schedule_retries():
    fixture = helpers.load_fixture('instalment_schedules')['create_with_schedule']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.instalment_schedules.create_with_schedule(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)

def test_502_instalment_schedules_create_with_schedule_retries():
    fixture = helpers.load_fixture('instalment_schedules')['create_with_schedule']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.instalment_schedules.create_with_schedule(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)
  

@responses.activate
def test_instalment_schedules_list():
    fixture = helpers.load_fixture('instalment_schedules')['list']
    helpers.stub_response(fixture)
    response = helpers.client.instalment_schedules.list(*fixture['url_params'])
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, list_response.ListResponse)
    assert_is_instance(response.records[0], resources.InstalmentSchedule)

    assert_equal(response.before, fixture['body']['meta']['cursors']['before'])
    assert_equal(response.after, fixture['body']['meta']['cursors']['after'])
    assert_is_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal([r.created_at for r in response.records],
                 [b.get('created_at') for b in body])
    assert_equal([r.currency for r in response.records],
                 [b.get('currency') for b in body])
    assert_equal([r.id for r in response.records],
                 [b.get('id') for b in body])
    assert_equal([r.metadata for r in response.records],
                 [b.get('metadata') for b in body])
    assert_equal([r.name for r in response.records],
                 [b.get('name') for b in body])
    assert_equal([r.payment_errors for r in response.records],
                 [b.get('payment_errors') for b in body])
    assert_equal([r.status for r in response.records],
                 [b.get('status') for b in body])
    assert_equal([r.total_amount for r in response.records],
                 [b.get('total_amount') for b in body])

@responses.activate
def test_timeout_instalment_schedules_list_retries():
    fixture = helpers.load_fixture('instalment_schedules')['list']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.instalment_schedules.list(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, list_response.ListResponse)
    assert_is_instance(response.records[0], resources.InstalmentSchedule)

    assert_equal(response.before, fixture['body']['meta']['cursors']['before'])
    assert_equal(response.after, fixture['body']['meta']['cursors']['after'])

def test_502_instalment_schedules_list_retries():
    fixture = helpers.load_fixture('instalment_schedules')['list']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.instalment_schedules.list(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, list_response.ListResponse)
    assert_is_instance(response.records[0], resources.InstalmentSchedule)

    assert_equal(response.before, fixture['body']['meta']['cursors']['before'])
    assert_equal(response.after, fixture['body']['meta']['cursors']['after'])

@responses.activate
def test_instalment_schedules_all():
    fixture = helpers.load_fixture('instalment_schedules')['list']

    def callback(request):
        if 'after=123' in request.url:
          fixture['body']['meta']['cursors']['after'] = None
        else:
          fixture['body']['meta']['cursors']['after'] = '123'
        return [200, {}, json.dumps(fixture['body'])]

    url = 'http://example.com' + fixture['path_template']
    responses.add_callback(fixture['method'], url, callback)

    all_records = list(helpers.client.instalment_schedules.all())
    assert_equal(len(all_records), len(fixture['body']['instalment_schedules']) * 2)
    for record in all_records:
      assert_is_instance(record, resources.InstalmentSchedule)
    
  

@responses.activate
def test_instalment_schedules_get():
    fixture = helpers.load_fixture('instalment_schedules')['get']
    helpers.stub_response(fixture)
    response = helpers.client.instalment_schedules.get(*fixture['url_params'])
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)
    assert_is_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.currency, body.get('currency'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.name, body.get('name'))
    assert_equal(response.payment_errors, body.get('payment_errors'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.total_amount, body.get('total_amount'))
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.mandate,
                 body.get('links')['mandate'])
    assert_equal(response.links.payments,
                 body.get('links')['payments'])

@responses.activate
def test_timeout_instalment_schedules_get_retries():
    fixture = helpers.load_fixture('instalment_schedules')['get']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.instalment_schedules.get(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)

def test_502_instalment_schedules_get_retries():
    fixture = helpers.load_fixture('instalment_schedules')['get']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.instalment_schedules.get(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)
  

@responses.activate
def test_instalment_schedules_update():
    fixture = helpers.load_fixture('instalment_schedules')['update']
    helpers.stub_response(fixture)
    response = helpers.client.instalment_schedules.update(*fixture['url_params'])
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)
    assert_is_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.currency, body.get('currency'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.name, body.get('name'))
    assert_equal(response.payment_errors, body.get('payment_errors'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.total_amount, body.get('total_amount'))
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.mandate,
                 body.get('links')['mandate'])
    assert_equal(response.links.payments,
                 body.get('links')['payments'])

@responses.activate
def test_timeout_instalment_schedules_update_retries():
    fixture = helpers.load_fixture('instalment_schedules')['update']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.instalment_schedules.update(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)

def test_502_instalment_schedules_update_retries():
    fixture = helpers.load_fixture('instalment_schedules')['update']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.instalment_schedules.update(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)
  

@responses.activate
def test_instalment_schedules_cancel():
    fixture = helpers.load_fixture('instalment_schedules')['cancel']
    helpers.stub_response(fixture)
    response = helpers.client.instalment_schedules.cancel(*fixture['url_params'])
    body = fixture['body']['instalment_schedules']

    assert_is_instance(response, resources.InstalmentSchedule)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.currency, body.get('currency'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.name, body.get('name'))
    assert_equal(response.payment_errors, body.get('payment_errors'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.total_amount, body.get('total_amount'))
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.mandate,
                 body.get('links')['mandate'])
    assert_equal(response.links.payments,
                 body.get('links')['payments'])

def test_timeout_instalment_schedules_cancel_doesnt_retry():
    fixture = helpers.load_fixture('instalment_schedules')['cancel']
    with helpers.stub_timeout(fixture) as rsps:
      with assert_raises(requests.ConnectTimeout):
        response = helpers.client.instalment_schedules.cancel(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))

def test_502_instalment_schedules_cancel_doesnt_retry():
    fixture = helpers.load_fixture('instalment_schedules')['cancel']
    with helpers.stub_502(fixture) as rsps:
      with assert_raises(MalformedResponseError):
        response = helpers.client.instalment_schedules.cancel(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

class Event(object):
    """A thin wrapper around a event, providing easy access to its
    attributes.

    Example:
      event = client.events.get()
      event.id
    """

    def __init__(self, attributes, api_response):
        self.attributes = attributes
        self.api_response = api_response

    @property
    def action(self):
        return self.attributes.get('action')
  

    @property
    def created_at(self):
        return self.attributes.get('created_at')
  

    @property
    def customer_notifications(self):
        return self.attributes.get('customer_notifications')
  

    @property
    def details(self):
        return self.Details(self.attributes.get('details'))
  

    @property
    def id(self):
        return self.attributes.get('id')
  

    @property
    def links(self):
        return self.Links(self.attributes.get('links'))
  

    @property
    def metadata(self):
        return self.attributes.get('metadata')
  

    @property
    def resource_type(self):
        return self.attributes.get('resource_type')
  


  

  

  

  
    class Details(object):
        """Wrapper for the response's 'details' attribute."""

        def __init__(self, attributes):
            self.attributes = attributes
    
        @property
        def bank_account_id(self):
            return self.attributes.get('bank_account_id')
    
        @property
        def cause(self):
            return self.attributes.get('cause')
    
        @property
        def currency(self):
            return self.attributes.get('currency')
    
        @property
        def description(self):
            return self.attributes.get('description')
    
        @property
        def not_retried_reason(self):
            return self.attributes.get('not_retried_reason')
    
        @property
        def origin(self):
            return self.attributes.get('origin')
    
        @property
        def _property(self):
            return self.attributes.get('property')
    
        @property
        def reason_code(self):
            return self.attributes.get('reason_code')
    
        @property
        def scheme(self):
            return self.attributes.get('scheme')
    
        @property
        def will_attempt_retry(self):
            return self.attributes.get('will_attempt_retry')
    
  

  

  
    class Links(object):
        """Wrapper for the response's 'links' attribute."""

        def __init__(self, attributes):
            self.attributes = attributes
    
        @property
        def bank_authorisation(self):
            return self.attributes.get('bank_authorisation')
    
        @property
        def billing_request(self):
            return self.attributes.get('billing_request')
    
        @property
        def billing_request_flow(self):
            return self.attributes.get('billing_request_flow')
    
        @property
        def creditor(self):
            return self.attributes.get('creditor')
    
        @property
        def customer(self):
            return self.attributes.get('customer')
    
        @property
        def customer_bank_account(self):
            return self.attributes.get('customer_bank_account')
    
        @property
        def instalment_schedule(self):
            return self.attributes.get('instalment_schedule')
    
        @property
        def mandate(self):
            return self.attributes.get('mandate')
    
        @property
        def mandate_request_mandate(self):
            return self.attributes.get('mandate_request_mandate')
    
        @property
        def new_customer_bank_account(self):
            return self.attributes.get('new_customer_bank_account')
    
        @property
        def new_mandate(self):
            return self.attributes.get('new_mandate')
    
        @property
        def organisation(self):
            return self.attributes.get('organisation')
    
        @property
        def parent_event(self):
            return self.attributes.get('parent_event')
    
        @property
        def payer_authorisation(self):
            return self.attributes.get('payer_authorisation')
    
        @property
        def payment(self):
            return self.attributes.get('payment')
    
        @property
        def payment_request_payment(self):
            return self.attributes.get('payment_request_payment')
    
        @property
        def payout(self):
            return self.attributes.get('payout')
    
        @property
        def previous_customer_bank_account(self):
            return self.attributes.get('previous_customer_bank_account')
    
        @property
        def refund(self):
            return self.attributes.get('refund')
    
        @property
        def subscription(self):
            return self.attributes.get('subscription')
    
  

  

  




# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class BankAuthorisationsService(base_service.BaseService):
    """Service class that provides access to the bank_authorisations
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.BankAuthorisation
    RESOURCE_NAME = 'bank_authorisations'


    def get(self,identity,params=None, headers=None):
        """Get a Bank Authorisation..

        Fetches a bank authorisation

        Args:
              identity (string): Unique identifier, beginning with "BAU".
              params (dict, optional): Query string parameters.

        Returns:
              BankAuthorisation
        """
        path = self._sub_url_params('/bank_authorisations/:identity', {
          
            'identity': identity,
          })
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def create(self,params=None, headers=None):
        """Create a Bank Authorisation.

        Create a Bank Authorisation.

        Args:
              params (dict, optional): Request body.

        Returns:
              BankAuthorisation
        """
        path = '/bank_authorisations'
        
        if params is not None:
            params = {self._envelope_key(): params}

        try:
          response = self._perform_request('POST', path, params, headers,
                                            retry_failures=True)
        except errors.IdempotentCreationConflictError as err:
          if self.raise_on_idempotency_conflict:
            raise err
          return self.get(identity=err.conflicting_resource_id,
                          params=params,
                          headers=headers)
        return self._resource_for(response)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

import json

import requests
import responses
from nose.tools import (
  assert_equal,
  assert_is_instance,
  assert_is_none,
  assert_is_not_none,
  assert_not_equal,
  assert_raises
)

from gocardless_pro.errors import MalformedResponseError
from gocardless_pro import resources
from gocardless_pro import list_response

from .. import helpers
  

@responses.activate
def test_scenario_simulators_run():
    fixture = helpers.load_fixture('scenario_simulators')['run']
    helpers.stub_response(fixture)
    response = helpers.client.scenario_simulators.run(*fixture['url_params'])
    body = fixture['body']['scenario_simulators']

    assert_is_instance(response, resources.ScenarioSimulator)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.id, body.get('id'))

def test_timeout_scenario_simulators_run_doesnt_retry():
    fixture = helpers.load_fixture('scenario_simulators')['run']
    with helpers.stub_timeout(fixture) as rsps:
      with assert_raises(requests.ConnectTimeout):
        response = helpers.client.scenario_simulators.run(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))

def test_502_scenario_simulators_run_doesnt_retry():
    fixture = helpers.load_fixture('scenario_simulators')['run']
    with helpers.stub_502(fixture) as rsps:
      with assert_raises(MalformedResponseError):
        response = helpers.client.scenario_simulators.run(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class CreditorsService(base_service.BaseService):
    """Service class that provides access to the creditors
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.Creditor
    RESOURCE_NAME = 'creditors'


    def create(self,params=None, headers=None):
        """Create a creditor.

        Creates a new creditor.

        Args:
              params (dict, optional): Request body.

        Returns:
              Creditor
        """
        path = '/creditors'
        
        if params is not None:
            params = {self._envelope_key(): params}

        try:
          response = self._perform_request('POST', path, params, headers,
                                            retry_failures=True)
        except errors.IdempotentCreationConflictError as err:
          if self.raise_on_idempotency_conflict:
            raise err
          return self.get(identity=err.conflicting_resource_id,
                          params=params,
                          headers=headers)
        return self._resource_for(response)
  

    def list(self,params=None, headers=None):
        """List creditors.

        Returns a [cursor-paginated](#api-usage-cursor-pagination) list of your
        creditors.

        Args:
              params (dict, optional): Query string parameters.

        Returns:
              ListResponse of Creditor instances
        """
        path = '/creditors'
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)

    def all(self, params=None):
        if params is None:
            params = {}
        return Paginator(self, params)
    
  

    def get(self,identity,params=None, headers=None):
        """Get a single creditor.

        Retrieves the details of an existing creditor.

        Args:
              identity (string): Unique identifier, beginning with "CR".
              params (dict, optional): Query string parameters.

        Returns:
              Creditor
        """
        path = self._sub_url_params('/creditors/:identity', {
          
            'identity': identity,
          })
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def update(self,identity,params=None, headers=None):
        """Update a creditor.

        Updates a creditor object. Supports all of the fields supported when
        creating a creditor.

        Args:
              identity (string): Unique identifier, beginning with "CR".
              params (dict, optional): Request body.

        Returns:
              Creditor
        """
        path = self._sub_url_params('/creditors/:identity', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {self._envelope_key(): params}

        response = self._perform_request('PUT', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

class BillingRequestTemplate(object):
    """A thin wrapper around a billing_request_template, providing easy access to its
    attributes.

    Example:
      billing_request_template = client.billing_request_templates.get()
      billing_request_template.id
    """

    def __init__(self, attributes, api_response):
        self.attributes = attributes
        self.api_response = api_response

    @property
    def authorisation_url(self):
        return self.attributes.get('authorisation_url')
  

    @property
    def created_at(self):
        return self.attributes.get('created_at')
  

    @property
    def id(self):
        return self.attributes.get('id')
  

    @property
    def mandate_request_currency(self):
        return self.attributes.get('mandate_request_currency')
  

    @property
    def mandate_request_metadata(self):
        return self.attributes.get('mandate_request_metadata')
  

    @property
    def mandate_request_scheme(self):
        return self.attributes.get('mandate_request_scheme')
  

    @property
    def mandate_request_verify(self):
        return self.attributes.get('mandate_request_verify')
  

    @property
    def metadata(self):
        return self.attributes.get('metadata')
  

    @property
    def name(self):
        return self.attributes.get('name')
  

    @property
    def payment_request_amount(self):
        return self.attributes.get('payment_request_amount')
  

    @property
    def payment_request_currency(self):
        return self.attributes.get('payment_request_currency')
  

    @property
    def payment_request_description(self):
        return self.attributes.get('payment_request_description')
  

    @property
    def payment_request_metadata(self):
        return self.attributes.get('payment_request_metadata')
  

    @property
    def payment_request_scheme(self):
        return self.attributes.get('payment_request_scheme')
  

    @property
    def redirect_uri(self):
        return self.attributes.get('redirect_uri')
  

    @property
    def updated_at(self):
        return self.attributes.get('updated_at')
  


  

  

  

  

  

  

  

  

  

  

  

  

  

  

  

  




# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class MandateImportsService(base_service.BaseService):
    """Service class that provides access to the mandate_imports
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.MandateImport
    RESOURCE_NAME = 'mandate_imports'


    def create(self,params=None, headers=None):
        """Create a new mandate import.

        Mandate imports are first created, before mandates are added
        one-at-a-time, so
        this endpoint merely signals the start of the import process. Once
        you've finished
        adding entries to an import, you should
        [submit](#mandate-imports-submit-a-mandate-import) it.

        Args:
              params (dict, optional): Request body.

        Returns:
              MandateImport
        """
        path = '/mandate_imports'
        
        if params is not None:
            params = {self._envelope_key(): params}

        try:
          response = self._perform_request('POST', path, params, headers,
                                            retry_failures=True)
        except errors.IdempotentCreationConflictError as err:
          if self.raise_on_idempotency_conflict:
            raise err
          return self.get(identity=err.conflicting_resource_id,
                          params=params,
                          headers=headers)
        return self._resource_for(response)
  

    def get(self,identity,params=None, headers=None):
        """Get a mandate import.

        Returns a single mandate import.

        Args:
              identity (string): Unique identifier, beginning with "IM".
              params (dict, optional): Query string parameters.

        Returns:
              MandateImport
        """
        path = self._sub_url_params('/mandate_imports/:identity', {
          
            'identity': identity,
          })
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def submit(self,identity,params=None, headers=None):
        """Submit a mandate import.

        Submits the mandate import, which allows it to be processed by a member
        of the
        GoCardless team. Once the import has been submitted, it can no longer
        have entries
        added to it.
        
        In our sandbox environment, to aid development, we automatically
        process mandate
        imports approximately 10 seconds after they are submitted. This will
        allow you to
        test both the "submitted" response and wait for the webhook to confirm
        the
        processing has begun.

        Args:
              identity (string): Unique identifier, beginning with "IM".
              params (dict, optional): Request body.

        Returns:
              MandateImport
        """
        path = self._sub_url_params('/mandate_imports/:identity/actions/submit', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def cancel(self,identity,params=None, headers=None):
        """Cancel a mandate import.

        Cancels the mandate import, which aborts the import process and stops
        the mandates
        being set up in GoCardless. Once the import has been cancelled, it can
        no longer have
        entries added to it. Mandate imports which have already been submitted
        or processed
        cannot be cancelled.

        Args:
              identity (string): Unique identifier, beginning with "IM".
              params (dict, optional): Request body.

        Returns:
              MandateImport
        """
        path = self._sub_url_params('/mandate_imports/:identity/actions/cancel', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

class BillingRequestFlow(object):
    """A thin wrapper around a billing_request_flow, providing easy access to its
    attributes.

    Example:
      billing_request_flow = client.billing_request_flows.get()
      billing_request_flow.id
    """

    def __init__(self, attributes, api_response):
        self.attributes = attributes
        self.api_response = api_response

    @property
    def authorisation_url(self):
        return self.attributes.get('authorisation_url')
  

    @property
    def auto_fulfil(self):
        return self.attributes.get('auto_fulfil')
  

    @property
    def created_at(self):
        return self.attributes.get('created_at')
  

    @property
    def exit_uri(self):
        return self.attributes.get('exit_uri')
  

    @property
    def expires_at(self):
        return self.attributes.get('expires_at')
  

    @property
    def id(self):
        return self.attributes.get('id')
  

    @property
    def links(self):
        return self.Links(self.attributes.get('links'))
  

    @property
    def lock_bank_account(self):
        return self.attributes.get('lock_bank_account')
  

    @property
    def lock_customer_details(self):
        return self.attributes.get('lock_customer_details')
  

    @property
    def redirect_uri(self):
        return self.attributes.get('redirect_uri')
  

    @property
    def session_token(self):
        return self.attributes.get('session_token')
  

    @property
    def show_redirect_buttons(self):
        return self.attributes.get('show_redirect_buttons')
  


  

  

  

  

  

  

  
    class Links(object):
        """Wrapper for the response's 'links' attribute."""

        def __init__(self, attributes):
            self.attributes = attributes
    
        @property
        def billing_request(self):
            return self.attributes.get('billing_request')
    
  

  

  

  

  

  




# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

import re
import time
from requests import Timeout, ConnectionError
from uuid import uuid4

from .. import list_response
from ..api_response import ApiResponse
from ..errors import MalformedResponseError

class BaseService(object):
    """Base class for API service classes."""
    def __init__(self, api_client, max_network_retries=3, retry_delay_in_seconds=0.5, raise_on_idempotency_conflict=False):
        self._api_client = api_client
        self.max_network_retries = max_network_retries
        self.retry_delay_in_seconds = retry_delay_in_seconds
        self.raise_on_idempotency_conflict = raise_on_idempotency_conflict

    def _perform_request(self, method, path, params, headers=None, retry_failures=False):
        if method == 'POST':
            headers = self._inject_idempotency_key(headers)

        if retry_failures:
            for retries_left in range(self.max_network_retries-1, -1, -1):
                try:
                    return self._attempt_request(method, path, params, headers)
                except (Timeout, ConnectionError, MalformedResponseError) as err:
                    if retries_left > 0:
                        time.sleep(self.retry_delay_in_seconds)
                    else:
                        raise err
        else:
            return self._attempt_request(method, path, params, headers)

    def _attempt_request(self, method, path, params, headers):
        if method == 'GET':
            return self._api_client.get(path, params=params, headers=headers)

        if method == 'POST':
            return self._api_client.post(path, body=params, headers=headers)

        if method == 'PUT':
            return self._api_client.put(path, body=params, headers=headers)

        if method == 'DELETE':
            return self._api_client.delete(path, body=params, headers=headers)

        raise ValueError('Invalid method "{}"'.format(method))


    def _inject_idempotency_key(self, headers):
        headers = headers or {}
        if 'Idempotency-Key' not in headers:
            headers['Idempotency-Key'] = str(uuid4())

        return headers

    def _envelope_key(self):
        return type(self).RESOURCE_NAME

    def _resource_for(self, response):
        api_response = ApiResponse(response)

        data = api_response.body[self._envelope_key()]
        klass = type(self).RESOURCE_CLASS
        if isinstance(data, dict):
            return klass(data, api_response)
        else:
            records = [klass(item, api_response) for item in data]
            return list_response.ListResponse(records, api_response)

    def _sub_url_params(self, url, params):
        return re.sub(r':(\w+)', lambda match: params[match.group(1)], url)



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

import json


class GoCardlessProError(Exception):
    """Base exception class for gocardless_pro errors."""


class ApiError(GoCardlessProError):
    """Base exception class for GoCardless API errors. API errors will result
    in one of this class' subclasses being raised.
    """

    def __init__(self, error):
        self.error = error

    def __str__(self):
        messages = [error['message'] for error in (self.errors or [])
                    if error['message'] != self.message]
        if messages:
            return '{} ({})'.format(self.message, ', '.join(messages))
        else:
            return self.message

    @property
    def type(self):
        return self.error.get('type')

    @property
    def code(self):
        return self.error.get('code')

    @property
    def errors(self):
        return self.error.get('errors')

    @property
    def documentation_url(self):
        return self.error.get('documentation_url')

    @property
    def message(self):
        return self.error.get('message')

    @property
    def request_id(self):
        return self.error.get('request_id')

    @classmethod
    def _exception_for_status(cls, status):
        if status == 401:
            return AuthenticationError
        elif status == 403:
            return PermissionsError
        elif status == 429:
            return RateLimitError
        else:
            return None

    @classmethod
    def _exception_for_error_type(cls, error_type, errors=[]):
        if error_type == 'validation_failed':
            return ValidationFailedError
        elif error_type == 'invalid_api_usage':
            return InvalidApiUsageError
        elif error_type == 'invalid_state':
            for error in errors:
                if error['reason'] == 'idempotent_creation_conflict':
                    return IdempotentCreationConflictError
            return InvalidStateError
        elif error_type == 'gocardless':
            return GoCardlessInternalError
        else:
            return None


    @classmethod
    def exception_for(cls, status, error_type, errors=[]):
        exception = cls._exception_for_status(status) or cls._exception_for_error_type(error_type, errors)
        if exception is not None:
            return exception
        raise GoCardlessProError('Invalid error type "{}"'.format(error_type))


class ValidationFailedError(ApiError):

    def __str__(self):
        if self.errors and 'field' in self.errors[0]:
            errors = ['{field} {message}'.format(**error) for error in self.errors]
            return '{} ({})'.format(self.message, ', '.join(errors))
        return super(ValidationFailedError, self).__str__()


class IdempotentCreationConflictError(ApiError):

    @property
    def conflicting_resource_id(self):
        for error in self.error['errors']:
            if 'conflicting_resource_id' in error.get('links', {}) and \
                    bool(error['links']['conflicting_resource_id']):  # Neither None nor ""
                return error['links']['conflicting_resource_id']
        else:
            raise ApiError("Idempotent Creation Conflict Error missing conflicting_resource_id")


class InvalidApiUsageError(ApiError):
    pass


class InvalidStateError(ApiError):
    pass


class GoCardlessInternalError(ApiError):
    pass


class AuthenticationError(ApiError):
    pass


class PermissionsError(ApiError):
    pass


class RateLimitError(ApiError):
    pass


class MalformedResponseError(GoCardlessProError):

    def __init__(self, message, response):
        super(MalformedResponseError, self).__init__(message)
        self.response = response


class InvalidSignatureError(GoCardlessProError):
    pass



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class BillingRequestTemplatesService(base_service.BaseService):
    """Service class that provides access to the billing_request_templates
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.BillingRequestTemplate
    RESOURCE_NAME = 'billing_request_templates'


    def list(self,params=None, headers=None):
        """List Billing Request Templates.

        Returns a [cursor-paginated](#api-usage-cursor-pagination) list of your
        Billing Request Templates.

        Args:
              params (dict, optional): Query string parameters.

        Returns:
              ListResponse of BillingRequestTemplate instances
        """
        path = '/billing_request_templates'
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)

    def all(self, params=None):
        if params is None:
            params = {}
        return Paginator(self, params)
    
  

    def get(self,identity,params=None, headers=None):
        """Get a single Billing Request Template.

        Fetches a Billing Request Template

        Args:
              identity (string): Unique identifier, beginning with "BRT".
              params (dict, optional): Query string parameters.

        Returns:
              BillingRequestTemplate
        """
        path = self._sub_url_params('/billing_request_templates/:identity', {
          
            'identity': identity,
          })
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def create(self,params=None, headers=None):
        """Create a Billing Request Template.

        

        Args:
              params (dict, optional): Request body.

        Returns:
              BillingRequestTemplate
        """
        path = '/billing_request_templates'
        
        if params is not None:
            params = {self._envelope_key(): params}

        try:
          response = self._perform_request('POST', path, params, headers,
                                            retry_failures=True)
        except errors.IdempotentCreationConflictError as err:
          if self.raise_on_idempotency_conflict:
            raise err
          return self.get(identity=err.conflicting_resource_id,
                          params=params,
                          headers=headers)
        return self._resource_for(response)
  

    def update(self,identity,params=None, headers=None):
        """Update a Billing Request Template.

        Updates a Billing Request Template, which will affect all future
        Billing Requests created by this template.

        Args:
              identity (string): Unique identifier, beginning with "BRQ".
              params (dict, optional): Request body.

        Returns:
              BillingRequestTemplate
        """
        path = self._sub_url_params('/billing_request_templates/:identity', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {self._envelope_key(): params}

        response = self._perform_request('PUT', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class MandateImportEntriesService(base_service.BaseService):
    """Service class that provides access to the mandate_import_entries
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.MandateImportEntry
    RESOURCE_NAME = 'mandate_import_entries'


    def create(self,params=None, headers=None):
        """Add a mandate import entry.

        For an existing [mandate import](#core-endpoints-mandate-imports), this
        endpoint can
        be used to add individual mandates to be imported into GoCardless.
        
        You can add no more than 30,000 rows to a single mandate import.
        If you attempt to go over this limit, the API will return a
        `record_limit_exceeded` error.

        Args:
              params (dict, optional): Request body.

        Returns:
              MandateImportEntry
        """
        path = '/mandate_import_entries'
        
        if params is not None:
            params = {self._envelope_key(): params}

        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def list(self,params=None, headers=None):
        """List all mandate import entries.

        For an existing mandate import, this endpoint lists all of the entries
        attached.
        
        After a mandate import has been submitted, you can use this endpoint to
        associate records
        in your system (using the `record_identifier` that you provided when
        creating the
        mandate import).
        

        Args:
              params (dict, optional): Query string parameters.

        Returns:
              ListResponse of MandateImportEntry instances
        """
        path = '/mandate_import_entries'
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)

    def all(self, params=None):
        if params is None:
            params = {}
        return Paginator(self, params)
    
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

class BillingRequest(object):
    """A thin wrapper around a billing_request, providing easy access to its
    attributes.

    Example:
      billing_request = client.billing_requests.get()
      billing_request.id
    """

    def __init__(self, attributes, api_response):
        self.attributes = attributes
        self.api_response = api_response

    @property
    def actions(self):
        return self.attributes.get('actions')
  

    @property
    def created_at(self):
        return self.attributes.get('created_at')
  

    @property
    def fallback_enabled(self):
        return self.attributes.get('fallback_enabled')
  

    @property
    def id(self):
        return self.attributes.get('id')
  

    @property
    def links(self):
        return self.Links(self.attributes.get('links'))
  

    @property
    def mandate_request(self):
        return self.MandateRequest(self.attributes.get('mandate_request'))
  

    @property
    def metadata(self):
        return self.attributes.get('metadata')
  

    @property
    def payment_request(self):
        return self.PaymentRequest(self.attributes.get('payment_request'))
  

    @property
    def resources(self):
        return self.Resources(self.attributes.get('resources'))
  

    @property
    def status(self):
        return self.attributes.get('status')
  


  

  

  

  

  
    class Links(object):
        """Wrapper for the response's 'links' attribute."""

        def __init__(self, attributes):
            self.attributes = attributes
    
        @property
        def bank_authorisation(self):
            return self.attributes.get('bank_authorisation')
    
        @property
        def creditor(self):
            return self.attributes.get('creditor')
    
        @property
        def customer(self):
            return self.attributes.get('customer')
    
        @property
        def customer_bank_account(self):
            return self.attributes.get('customer_bank_account')
    
        @property
        def customer_billing_detail(self):
            return self.attributes.get('customer_billing_detail')
    
        @property
        def mandate_request(self):
            return self.attributes.get('mandate_request')
    
        @property
        def mandate_request_mandate(self):
            return self.attributes.get('mandate_request_mandate')
    
        @property
        def payment_request(self):
            return self.attributes.get('payment_request')
    
        @property
        def payment_request_payment(self):
            return self.attributes.get('payment_request_payment')
    
  

  
    class MandateRequest(object):
        """Wrapper for the response's 'mandate_request' attribute."""

        def __init__(self, attributes):
            self.attributes = attributes
    
        @property
        def currency(self):
            return self.attributes.get('currency')
    
        @property
        def links(self):
            return self.attributes.get('links')
    
        @property
        def metadata(self):
            return self.attributes.get('metadata')
    
        @property
        def scheme(self):
            return self.attributes.get('scheme')
    
        @property
        def verify(self):
            return self.attributes.get('verify')
    
  

  

  
    class PaymentRequest(object):
        """Wrapper for the response's 'payment_request' attribute."""

        def __init__(self, attributes):
            self.attributes = attributes
    
        @property
        def amount(self):
            return self.attributes.get('amount')
    
        @property
        def app_fee(self):
            return self.attributes.get('app_fee')
    
        @property
        def currency(self):
            return self.attributes.get('currency')
    
        @property
        def description(self):
            return self.attributes.get('description')
    
        @property
        def links(self):
            return self.attributes.get('links')
    
        @property
        def metadata(self):
            return self.attributes.get('metadata')
    
        @property
        def scheme(self):
            return self.attributes.get('scheme')
    
  

  
    class Resources(object):
        """Wrapper for the response's 'resources' attribute."""

        def __init__(self, attributes):
            self.attributes = attributes
    
        @property
        def customer(self):
            return self.attributes.get('customer')
    
        @property
        def customer_bank_account(self):
            return self.attributes.get('customer_bank_account')
    
        @property
        def customer_billing_detail(self):
            return self.attributes.get('customer_billing_detail')
    
  

  




# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

class Block(object):
    """A thin wrapper around a block, providing easy access to its
    attributes.

    Example:
      block = client.blocks.get()
      block.id
    """

    def __init__(self, attributes, api_response):
        self.attributes = attributes
        self.api_response = api_response

    @property
    def active(self):
        return self.attributes.get('active')
  

    @property
    def block_type(self):
        return self.attributes.get('block_type')
  

    @property
    def created_at(self):
        return self.attributes.get('created_at')
  

    @property
    def id(self):
        return self.attributes.get('id')
  

    @property
    def reason_description(self):
        return self.attributes.get('reason_description')
  

    @property
    def reason_type(self):
        return self.attributes.get('reason_type')
  

    @property
    def resource_reference(self):
        return self.attributes.get('resource_reference')
  

    @property
    def updated_at(self):
        return self.attributes.get('updated_at')
  


  

  

  

  

  

  

  

  




# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class BillingRequestsService(base_service.BaseService):
    """Service class that provides access to the billing_requests
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.BillingRequest
    RESOURCE_NAME = 'billing_requests'


    def list(self,params=None, headers=None):
        """List Billing Requests.

        Returns a [cursor-paginated](#api-usage-cursor-pagination) list of your
        billing requests.

        Args:
              params (dict, optional): Query string parameters.

        Returns:
              ListResponse of BillingRequest instances
        """
        path = '/billing_requests'
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)

    def all(self, params=None):
        if params is None:
            params = {}
        return Paginator(self, params)
    
  

    def create(self,params=None, headers=None):
        """Create a Billing Request.

        

        Args:
              params (dict, optional): Request body.

        Returns:
              BillingRequest
        """
        path = '/billing_requests'
        
        if params is not None:
            params = {self._envelope_key(): params}

        try:
          response = self._perform_request('POST', path, params, headers,
                                            retry_failures=True)
        except errors.IdempotentCreationConflictError as err:
          if self.raise_on_idempotency_conflict:
            raise err
          return self.get(identity=err.conflicting_resource_id,
                          params=params,
                          headers=headers)
        return self._resource_for(response)
  

    def get(self,identity,params=None, headers=None):
        """Get a single Billing Request.

        Fetches a billing request

        Args:
              identity (string): Unique identifier, beginning with "BRQ".
              params (dict, optional): Query string parameters.

        Returns:
              BillingRequest
        """
        path = self._sub_url_params('/billing_requests/:identity', {
          
            'identity': identity,
          })
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def collect_customer_details(self,identity,params=None, headers=None):
        """Collect customer details for a Billing Request.

        If the billing request has a pending
        <code>collect_customer_details</code>
        action, this endpoint can be used to collect the details in order to
        complete it.
        
        The endpoint takes the same payload as Customers, but checks that the
        customer fields are populated correctly for the billing request scheme.
        
        Whatever is provided to this endpoint is used to update the referenced
        customer, and will take effect immediately after the request is
        successful.

        Args:
              identity (string): Unique identifier, beginning with "BRQ".
              params (dict, optional): Request body.

        Returns:
              BillingRequest
        """
        path = self._sub_url_params('/billing_requests/:identity/actions/collect_customer_details', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def collect_bank_account(self,identity,params=None, headers=None):
        """Collect bank account details for a Billing Request.

        If the billing request has a pending
        <code>collect_bank_account</code> action, this endpoint can be
        used to collect the details in order to complete it.
        
        The endpoint takes the same payload as Customer Bank Accounts, but
        check
        the bank account is valid for the billing request scheme before
        creating
        and attaching it.

        Args:
              identity (string): Unique identifier, beginning with "BRQ".
              params (dict, optional): Request body.

        Returns:
              BillingRequest
        """
        path = self._sub_url_params('/billing_requests/:identity/actions/collect_bank_account', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def fulfil(self,identity,params=None, headers=None):
        """Fulfil a Billing Request.

        If a billing request is ready to be fulfilled, call this endpoint to
        cause
        it to fulfil, executing the payment.

        Args:
              identity (string): Unique identifier, beginning with "BRQ".
              params (dict, optional): Request body.

        Returns:
              BillingRequest
        """
        path = self._sub_url_params('/billing_requests/:identity/actions/fulfil', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def confirm_payer_details(self,identity,params=None, headers=None):
        """Confirm the customer and bank account details.

        This is needed when you have a mandate request. As a scheme compliance
        rule we are required to
        allow the payer to crosscheck the details entered by them and confirm
        it.

        Args:
              identity (string): Unique identifier, beginning with "BRQ".
              params (dict, optional): Request body.

        Returns:
              BillingRequest
        """
        path = self._sub_url_params('/billing_requests/:identity/actions/confirm_payer_details', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def cancel(self,identity,params=None, headers=None):
        """Cancel a Billing Request.

        Immediately cancels a billing request, causing all billing request
        flows
        to expire.

        Args:
              identity (string): Unique identifier, beginning with "BRQ".
              params (dict, optional): Request body.

        Returns:
              BillingRequest
        """
        path = self._sub_url_params('/billing_requests/:identity/actions/cancel', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def notify(self,identity,params=None, headers=None):
        """Notify the customer of a Billing Request.

        Notifies the customer linked to the billing request, asking them to
        authorise it.
        Currently, the customer can only be notified by email.

        Args:
              identity (string): Unique identifier, beginning with "BRQ".
              params (dict, optional): Request body.

        Returns:
              BillingRequest
        """
        path = self._sub_url_params('/billing_requests/:identity/actions/notify', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def fallback(self,identity,params=None, headers=None):
        """Trigger fallback for a Billing Request.

        Triggers a fallback from the open-banking flow to direct debit. Note,
        the billing request must have fallback enabled.

        Args:
              identity (string): Unique identifier, beginning with "BRQ".
              params (dict, optional): Request body.

        Returns:
              BillingRequest
        """
        path = self._sub_url_params('/billing_requests/:identity/actions/fallback', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class CurrencyExchangeRatesService(base_service.BaseService):
    """Service class that provides access to the currency_exchange_rates
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.CurrencyExchangeRate
    RESOURCE_NAME = 'currency_exchange_rates'


    def list(self,params=None, headers=None):
        """List exchange rates.

        Returns a [cursor-paginated](#api-usage-cursor-pagination) list of all
        exchange rates.

        Args:
              params (dict, optional): Query string parameters.

        Returns:
              ListResponse of CurrencyExchangeRate instances
        """
        path = '/currency_exchange_rates'
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)

    def all(self, params=None):
        if params is None:
            params = {}
        return Paginator(self, params)
    
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

class ListResponse(object):

    def __init__(self, records, api_response):
        self.records = records
        self.api_response = api_response

    @property
    def before(self):
        return self.api_response.body['meta']['cursors']['before']

    @property
    def after(self):
        return self.api_response.body['meta']['cursors']['after']



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class SubscriptionsService(base_service.BaseService):
    """Service class that provides access to the subscriptions
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.Subscription
    RESOURCE_NAME = 'subscriptions'


    def create(self,params=None, headers=None):
        """Create a subscription.

        Creates a new subscription object

        Args:
              params (dict, optional): Request body.

        Returns:
              Subscription
        """
        path = '/subscriptions'
        
        if params is not None:
            params = {self._envelope_key(): params}

        try:
          response = self._perform_request('POST', path, params, headers,
                                            retry_failures=True)
        except errors.IdempotentCreationConflictError as err:
          if self.raise_on_idempotency_conflict:
            raise err
          return self.get(identity=err.conflicting_resource_id,
                          params=params,
                          headers=headers)
        return self._resource_for(response)
  

    def list(self,params=None, headers=None):
        """List subscriptions.

        Returns a [cursor-paginated](#api-usage-cursor-pagination) list of your
        subscriptions.

        Args:
              params (dict, optional): Query string parameters.

        Returns:
              ListResponse of Subscription instances
        """
        path = '/subscriptions'
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)

    def all(self, params=None):
        if params is None:
            params = {}
        return Paginator(self, params)
    
  

    def get(self,identity,params=None, headers=None):
        """Get a single subscription.

        Retrieves the details of a single subscription.

        Args:
              identity (string): Unique identifier, beginning with "SB".
              params (dict, optional): Query string parameters.

        Returns:
              Subscription
        """
        path = self._sub_url_params('/subscriptions/:identity', {
          
            'identity': identity,
          })
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def update(self,identity,params=None, headers=None):
        """Update a subscription.

        Updates a subscription object.
        
        This fails with:
        
        - `validation_failed` if invalid data is provided when attempting to
        update a subscription.
        
        - `subscription_not_active` if the subscription is no longer active.
        
        - `subscription_already_ended` if the subscription has taken all
        payments.
        
        - `mandate_payments_require_approval` if the amount is being changed
        and the mandate requires approval.
        
        - `number_of_subscription_amendments_exceeded` error if the
        subscription amount has already been changed 10 times.
        
        - `forbidden` if the amount is being changed, and the subscription was
        created by an app and you are not authenticated as that app, or if the
        subscription was not created by an app and you are authenticated as an
        app
        
        - `resource_created_by_another_app` if the app fee is being changed,
        and the subscription was created by an app other than the app you are
        authenticated as
        

        Args:
              identity (string): Unique identifier, beginning with "SB".
              params (dict, optional): Request body.

        Returns:
              Subscription
        """
        path = self._sub_url_params('/subscriptions/:identity', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {self._envelope_key(): params}

        response = self._perform_request('PUT', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def pause(self,identity,params=None, headers=None):
        """Pause a subscription.

        Pause a subscription object.
        No payments will be created until it is resumed.
        
        This can only be used when a subscription is collecting a fixed number
        of payments (created using `count`),
        when they continue forever (created without `count` or `end_date`) or
        the subscription is already paused for a number of cycles.
        
        When `pause_cycles` is omitted the subscription is paused until the
        [resume endpoint](#subscriptions-resume-a-subscription) is called.
        If the subscription is collecting a fixed number of payments,
        `end_date` will be set to `null`.
        When paused indefinitely, `upcoming_payments` will be empty.
        
        When `pause_cycles` is provided the subscription will be paused for the
        number of cycles requested.
        If the subscription is collecting a fixed number of payments,
        `end_date` will be set to a new value.
        When paused for a number of cycles, `upcoming_payments` will still
        contain the upcoming charge dates.
        
        This fails with:
        
        - `forbidden` if the subscription was created by an app and you are not
        authenticated as that app, or if the subscription was not created by an
        app and you are authenticated as an app
        
        - `validation_failed` if invalid data is provided when attempting to
        pause a subscription.
        
        - `subscription_paused_cannot_update_cycles` if the subscription is
        already paused for a number of cycles and the request provides a value
        for `pause_cycle`.
        
        - `subscription_cannot_be_paused` if the subscription cannot be paused.
        
        - `subscription_already_ended` if the subscription has taken all
        payments.
        
        - `pause_cycles_must_be_greater_than_or_equal_to` if the provided value
        for `pause_cycles` cannot be satisfied.
        

        Args:
              identity (string): Unique identifier, beginning with "SB".
              params (dict, optional): Request body.

        Returns:
              Subscription
        """
        path = self._sub_url_params('/subscriptions/:identity/actions/pause', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def resume(self,identity,params=None, headers=None):
        """Resume a subscription.

        Resume a subscription object.
        Payments will start to be created again based on the subscriptions
        recurrence rules.
        The `charge_date` on the next payment will be the same as the
        subscriptions `earliest_charge_date_after_resume`
        
        This fails with:
        
        - `forbidden` if the subscription was created by an app and you are not
        authenticated as that app, or if the subscription was not created by an
        app and you are authenticated as an app
        
        - `validation_failed` if invalid data is provided when attempting to
        resume a subscription.
        
        - `subscription_not_paused` if the subscription is not paused.
        

        Args:
              identity (string): Unique identifier, beginning with "SB".
              params (dict, optional): Request body.

        Returns:
              Subscription
        """
        path = self._sub_url_params('/subscriptions/:identity/actions/resume', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def cancel(self,identity,params=None, headers=None):
        """Cancel a subscription.

        Immediately cancels a subscription; no more payments will be created
        under it. Any metadata supplied to this endpoint will be stored on the
        payment cancellation event it causes.
        
        This will fail with a cancellation_failed error if the subscription is
        already cancelled or finished.

        Args:
              identity (string): Unique identifier, beginning with "SB".
              params (dict, optional): Request body.

        Returns:
              Subscription
        """
        path = self._sub_url_params('/subscriptions/:identity/actions/cancel', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

import json

import requests
import responses
from nose.tools import (
  assert_equal,
  assert_is_instance,
  assert_is_none,
  assert_is_not_none,
  assert_not_equal,
  assert_raises
)

from gocardless_pro.errors import MalformedResponseError
from gocardless_pro import resources
from gocardless_pro import list_response

from .. import helpers
  

@responses.activate
def test_billing_requests_list():
    fixture = helpers.load_fixture('billing_requests')['list']
    helpers.stub_response(fixture)
    response = helpers.client.billing_requests.list(*fixture['url_params'])
    body = fixture['body']['billing_requests']

    assert_is_instance(response, list_response.ListResponse)
    assert_is_instance(response.records[0], resources.BillingRequest)

    assert_equal(response.before, fixture['body']['meta']['cursors']['before'])
    assert_equal(response.after, fixture['body']['meta']['cursors']['after'])
    assert_is_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal([r.actions for r in response.records],
                 [b.get('actions') for b in body])
    assert_equal([r.created_at for r in response.records],
                 [b.get('created_at') for b in body])
    assert_equal([r.fallback_enabled for r in response.records],
                 [b.get('fallback_enabled') for b in body])
    assert_equal([r.id for r in response.records],
                 [b.get('id') for b in body])
    assert_equal([r.metadata for r in response.records],
                 [b.get('metadata') for b in body])
    assert_equal([r.status for r in response.records],
                 [b.get('status') for b in body])

@responses.activate
def test_timeout_billing_requests_list_retries():
    fixture = helpers.load_fixture('billing_requests')['list']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.billing_requests.list(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['billing_requests']

    assert_is_instance(response, list_response.ListResponse)
    assert_is_instance(response.records[0], resources.BillingRequest)

    assert_equal(response.before, fixture['body']['meta']['cursors']['before'])
    assert_equal(response.after, fixture['body']['meta']['cursors']['after'])

def test_502_billing_requests_list_retries():
    fixture = helpers.load_fixture('billing_requests')['list']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.billing_requests.list(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['billing_requests']

    assert_is_instance(response, list_response.ListResponse)
    assert_is_instance(response.records[0], resources.BillingRequest)

    assert_equal(response.before, fixture['body']['meta']['cursors']['before'])
    assert_equal(response.after, fixture['body']['meta']['cursors']['after'])

@responses.activate
def test_billing_requests_all():
    fixture = helpers.load_fixture('billing_requests')['list']

    def callback(request):
        if 'after=123' in request.url:
          fixture['body']['meta']['cursors']['after'] = None
        else:
          fixture['body']['meta']['cursors']['after'] = '123'
        return [200, {}, json.dumps(fixture['body'])]

    url = 'http://example.com' + fixture['path_template']
    responses.add_callback(fixture['method'], url, callback)

    all_records = list(helpers.client.billing_requests.all())
    assert_equal(len(all_records), len(fixture['body']['billing_requests']) * 2)
    for record in all_records:
      assert_is_instance(record, resources.BillingRequest)
    
  

@responses.activate
def test_billing_requests_create():
    fixture = helpers.load_fixture('billing_requests')['create']
    helpers.stub_response(fixture)
    response = helpers.client.billing_requests.create(*fixture['url_params'])
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.actions, body.get('actions'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.fallback_enabled, body.get('fallback_enabled'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.links.bank_authorisation,
                 body.get('links')['bank_authorisation'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.customer_billing_detail,
                 body.get('links')['customer_billing_detail'])
    assert_equal(response.links.mandate_request,
                 body.get('links')['mandate_request'])
    assert_equal(response.links.mandate_request_mandate,
                 body.get('links')['mandate_request_mandate'])
    assert_equal(response.links.payment_request,
                 body.get('links')['payment_request'])
    assert_equal(response.links.payment_request_payment,
                 body.get('links')['payment_request_payment'])
    assert_equal(response.mandate_request.currency,
                 body.get('mandate_request')['currency'])
    assert_equal(response.mandate_request.links,
                 body.get('mandate_request')['links'])
    assert_equal(response.mandate_request.metadata,
                 body.get('mandate_request')['metadata'])
    assert_equal(response.mandate_request.scheme,
                 body.get('mandate_request')['scheme'])
    assert_equal(response.mandate_request.verify,
                 body.get('mandate_request')['verify'])
    assert_equal(response.payment_request.amount,
                 body.get('payment_request')['amount'])
    assert_equal(response.payment_request.app_fee,
                 body.get('payment_request')['app_fee'])
    assert_equal(response.payment_request.currency,
                 body.get('payment_request')['currency'])
    assert_equal(response.payment_request.description,
                 body.get('payment_request')['description'])
    assert_equal(response.payment_request.links,
                 body.get('payment_request')['links'])
    assert_equal(response.payment_request.metadata,
                 body.get('payment_request')['metadata'])
    assert_equal(response.payment_request.scheme,
                 body.get('payment_request')['scheme'])
    assert_equal(response.resources.customer,
                 body.get('resources')['customer'])
    assert_equal(response.resources.customer_bank_account,
                 body.get('resources')['customer_bank_account'])
    assert_equal(response.resources.customer_billing_detail,
                 body.get('resources')['customer_billing_detail'])

@responses.activate
def test_billing_requests_create_new_idempotency_key_for_each_call():
    fixture = helpers.load_fixture('billing_requests')['create']
    helpers.stub_response(fixture)
    helpers.client.billing_requests.create(*fixture['url_params'])
    helpers.client.billing_requests.create(*fixture['url_params'])
    assert_not_equal(responses.calls[0].request.headers.get('Idempotency-Key'),
                     responses.calls[1].request.headers.get('Idempotency-Key'))

def test_timeout_billing_requests_create_idempotency_conflict():
    create_fixture = helpers.load_fixture('billing_requests')['create']
    get_fixture = helpers.load_fixture('billing_requests')['get']
    with helpers.stub_timeout_then_idempotency_conflict(create_fixture, get_fixture) as rsps:
      response = helpers.client.billing_requests.create(*create_fixture['url_params'])
      assert_equal(2, len(rsps.calls))

    assert_is_instance(response, resources.BillingRequest)

@responses.activate
def test_timeout_billing_requests_create_retries():
    fixture = helpers.load_fixture('billing_requests')['create']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.billing_requests.create(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)

def test_502_billing_requests_create_retries():
    fixture = helpers.load_fixture('billing_requests')['create']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.billing_requests.create(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)
  

@responses.activate
def test_billing_requests_get():
    fixture = helpers.load_fixture('billing_requests')['get']
    helpers.stub_response(fixture)
    response = helpers.client.billing_requests.get(*fixture['url_params'])
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)
    assert_is_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.actions, body.get('actions'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.fallback_enabled, body.get('fallback_enabled'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.links.bank_authorisation,
                 body.get('links')['bank_authorisation'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.customer_billing_detail,
                 body.get('links')['customer_billing_detail'])
    assert_equal(response.links.mandate_request,
                 body.get('links')['mandate_request'])
    assert_equal(response.links.mandate_request_mandate,
                 body.get('links')['mandate_request_mandate'])
    assert_equal(response.links.payment_request,
                 body.get('links')['payment_request'])
    assert_equal(response.links.payment_request_payment,
                 body.get('links')['payment_request_payment'])
    assert_equal(response.mandate_request.currency,
                 body.get('mandate_request')['currency'])
    assert_equal(response.mandate_request.links,
                 body.get('mandate_request')['links'])
    assert_equal(response.mandate_request.metadata,
                 body.get('mandate_request')['metadata'])
    assert_equal(response.mandate_request.scheme,
                 body.get('mandate_request')['scheme'])
    assert_equal(response.mandate_request.verify,
                 body.get('mandate_request')['verify'])
    assert_equal(response.payment_request.amount,
                 body.get('payment_request')['amount'])
    assert_equal(response.payment_request.app_fee,
                 body.get('payment_request')['app_fee'])
    assert_equal(response.payment_request.currency,
                 body.get('payment_request')['currency'])
    assert_equal(response.payment_request.description,
                 body.get('payment_request')['description'])
    assert_equal(response.payment_request.links,
                 body.get('payment_request')['links'])
    assert_equal(response.payment_request.metadata,
                 body.get('payment_request')['metadata'])
    assert_equal(response.payment_request.scheme,
                 body.get('payment_request')['scheme'])
    assert_equal(response.resources.customer,
                 body.get('resources')['customer'])
    assert_equal(response.resources.customer_bank_account,
                 body.get('resources')['customer_bank_account'])
    assert_equal(response.resources.customer_billing_detail,
                 body.get('resources')['customer_billing_detail'])

@responses.activate
def test_timeout_billing_requests_get_retries():
    fixture = helpers.load_fixture('billing_requests')['get']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.billing_requests.get(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)

def test_502_billing_requests_get_retries():
    fixture = helpers.load_fixture('billing_requests')['get']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.billing_requests.get(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)
  

@responses.activate
def test_billing_requests_collect_customer_details():
    fixture = helpers.load_fixture('billing_requests')['collect_customer_details']
    helpers.stub_response(fixture)
    response = helpers.client.billing_requests.collect_customer_details(*fixture['url_params'])
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.actions, body.get('actions'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.fallback_enabled, body.get('fallback_enabled'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.links.bank_authorisation,
                 body.get('links')['bank_authorisation'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.customer_billing_detail,
                 body.get('links')['customer_billing_detail'])
    assert_equal(response.links.mandate_request,
                 body.get('links')['mandate_request'])
    assert_equal(response.links.mandate_request_mandate,
                 body.get('links')['mandate_request_mandate'])
    assert_equal(response.links.payment_request,
                 body.get('links')['payment_request'])
    assert_equal(response.links.payment_request_payment,
                 body.get('links')['payment_request_payment'])
    assert_equal(response.mandate_request.currency,
                 body.get('mandate_request')['currency'])
    assert_equal(response.mandate_request.links,
                 body.get('mandate_request')['links'])
    assert_equal(response.mandate_request.metadata,
                 body.get('mandate_request')['metadata'])
    assert_equal(response.mandate_request.scheme,
                 body.get('mandate_request')['scheme'])
    assert_equal(response.mandate_request.verify,
                 body.get('mandate_request')['verify'])
    assert_equal(response.payment_request.amount,
                 body.get('payment_request')['amount'])
    assert_equal(response.payment_request.app_fee,
                 body.get('payment_request')['app_fee'])
    assert_equal(response.payment_request.currency,
                 body.get('payment_request')['currency'])
    assert_equal(response.payment_request.description,
                 body.get('payment_request')['description'])
    assert_equal(response.payment_request.links,
                 body.get('payment_request')['links'])
    assert_equal(response.payment_request.metadata,
                 body.get('payment_request')['metadata'])
    assert_equal(response.payment_request.scheme,
                 body.get('payment_request')['scheme'])
    assert_equal(response.resources.customer,
                 body.get('resources')['customer'])
    assert_equal(response.resources.customer_bank_account,
                 body.get('resources')['customer_bank_account'])
    assert_equal(response.resources.customer_billing_detail,
                 body.get('resources')['customer_billing_detail'])

def test_timeout_billing_requests_collect_customer_details_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['collect_customer_details']
    with helpers.stub_timeout(fixture) as rsps:
      with assert_raises(requests.ConnectTimeout):
        response = helpers.client.billing_requests.collect_customer_details(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))

def test_502_billing_requests_collect_customer_details_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['collect_customer_details']
    with helpers.stub_502(fixture) as rsps:
      with assert_raises(MalformedResponseError):
        response = helpers.client.billing_requests.collect_customer_details(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))
  

@responses.activate
def test_billing_requests_collect_bank_account():
    fixture = helpers.load_fixture('billing_requests')['collect_bank_account']
    helpers.stub_response(fixture)
    response = helpers.client.billing_requests.collect_bank_account(*fixture['url_params'])
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.actions, body.get('actions'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.fallback_enabled, body.get('fallback_enabled'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.links.bank_authorisation,
                 body.get('links')['bank_authorisation'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.customer_billing_detail,
                 body.get('links')['customer_billing_detail'])
    assert_equal(response.links.mandate_request,
                 body.get('links')['mandate_request'])
    assert_equal(response.links.mandate_request_mandate,
                 body.get('links')['mandate_request_mandate'])
    assert_equal(response.links.payment_request,
                 body.get('links')['payment_request'])
    assert_equal(response.links.payment_request_payment,
                 body.get('links')['payment_request_payment'])
    assert_equal(response.mandate_request.currency,
                 body.get('mandate_request')['currency'])
    assert_equal(response.mandate_request.links,
                 body.get('mandate_request')['links'])
    assert_equal(response.mandate_request.metadata,
                 body.get('mandate_request')['metadata'])
    assert_equal(response.mandate_request.scheme,
                 body.get('mandate_request')['scheme'])
    assert_equal(response.mandate_request.verify,
                 body.get('mandate_request')['verify'])
    assert_equal(response.payment_request.amount,
                 body.get('payment_request')['amount'])
    assert_equal(response.payment_request.app_fee,
                 body.get('payment_request')['app_fee'])
    assert_equal(response.payment_request.currency,
                 body.get('payment_request')['currency'])
    assert_equal(response.payment_request.description,
                 body.get('payment_request')['description'])
    assert_equal(response.payment_request.links,
                 body.get('payment_request')['links'])
    assert_equal(response.payment_request.metadata,
                 body.get('payment_request')['metadata'])
    assert_equal(response.payment_request.scheme,
                 body.get('payment_request')['scheme'])
    assert_equal(response.resources.customer,
                 body.get('resources')['customer'])
    assert_equal(response.resources.customer_bank_account,
                 body.get('resources')['customer_bank_account'])
    assert_equal(response.resources.customer_billing_detail,
                 body.get('resources')['customer_billing_detail'])

def test_timeout_billing_requests_collect_bank_account_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['collect_bank_account']
    with helpers.stub_timeout(fixture) as rsps:
      with assert_raises(requests.ConnectTimeout):
        response = helpers.client.billing_requests.collect_bank_account(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))

def test_502_billing_requests_collect_bank_account_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['collect_bank_account']
    with helpers.stub_502(fixture) as rsps:
      with assert_raises(MalformedResponseError):
        response = helpers.client.billing_requests.collect_bank_account(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))
  

@responses.activate
def test_billing_requests_fulfil():
    fixture = helpers.load_fixture('billing_requests')['fulfil']
    helpers.stub_response(fixture)
    response = helpers.client.billing_requests.fulfil(*fixture['url_params'])
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.actions, body.get('actions'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.fallback_enabled, body.get('fallback_enabled'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.links.bank_authorisation,
                 body.get('links')['bank_authorisation'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.customer_billing_detail,
                 body.get('links')['customer_billing_detail'])
    assert_equal(response.links.mandate_request,
                 body.get('links')['mandate_request'])
    assert_equal(response.links.mandate_request_mandate,
                 body.get('links')['mandate_request_mandate'])
    assert_equal(response.links.payment_request,
                 body.get('links')['payment_request'])
    assert_equal(response.links.payment_request_payment,
                 body.get('links')['payment_request_payment'])
    assert_equal(response.mandate_request.currency,
                 body.get('mandate_request')['currency'])
    assert_equal(response.mandate_request.links,
                 body.get('mandate_request')['links'])
    assert_equal(response.mandate_request.metadata,
                 body.get('mandate_request')['metadata'])
    assert_equal(response.mandate_request.scheme,
                 body.get('mandate_request')['scheme'])
    assert_equal(response.mandate_request.verify,
                 body.get('mandate_request')['verify'])
    assert_equal(response.payment_request.amount,
                 body.get('payment_request')['amount'])
    assert_equal(response.payment_request.app_fee,
                 body.get('payment_request')['app_fee'])
    assert_equal(response.payment_request.currency,
                 body.get('payment_request')['currency'])
    assert_equal(response.payment_request.description,
                 body.get('payment_request')['description'])
    assert_equal(response.payment_request.links,
                 body.get('payment_request')['links'])
    assert_equal(response.payment_request.metadata,
                 body.get('payment_request')['metadata'])
    assert_equal(response.payment_request.scheme,
                 body.get('payment_request')['scheme'])
    assert_equal(response.resources.customer,
                 body.get('resources')['customer'])
    assert_equal(response.resources.customer_bank_account,
                 body.get('resources')['customer_bank_account'])
    assert_equal(response.resources.customer_billing_detail,
                 body.get('resources')['customer_billing_detail'])

def test_timeout_billing_requests_fulfil_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['fulfil']
    with helpers.stub_timeout(fixture) as rsps:
      with assert_raises(requests.ConnectTimeout):
        response = helpers.client.billing_requests.fulfil(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))

def test_502_billing_requests_fulfil_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['fulfil']
    with helpers.stub_502(fixture) as rsps:
      with assert_raises(MalformedResponseError):
        response = helpers.client.billing_requests.fulfil(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))
  

@responses.activate
def test_billing_requests_confirm_payer_details():
    fixture = helpers.load_fixture('billing_requests')['confirm_payer_details']
    helpers.stub_response(fixture)
    response = helpers.client.billing_requests.confirm_payer_details(*fixture['url_params'])
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.actions, body.get('actions'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.fallback_enabled, body.get('fallback_enabled'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.links.bank_authorisation,
                 body.get('links')['bank_authorisation'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.customer_billing_detail,
                 body.get('links')['customer_billing_detail'])
    assert_equal(response.links.mandate_request,
                 body.get('links')['mandate_request'])
    assert_equal(response.links.mandate_request_mandate,
                 body.get('links')['mandate_request_mandate'])
    assert_equal(response.links.payment_request,
                 body.get('links')['payment_request'])
    assert_equal(response.links.payment_request_payment,
                 body.get('links')['payment_request_payment'])
    assert_equal(response.mandate_request.currency,
                 body.get('mandate_request')['currency'])
    assert_equal(response.mandate_request.links,
                 body.get('mandate_request')['links'])
    assert_equal(response.mandate_request.metadata,
                 body.get('mandate_request')['metadata'])
    assert_equal(response.mandate_request.scheme,
                 body.get('mandate_request')['scheme'])
    assert_equal(response.mandate_request.verify,
                 body.get('mandate_request')['verify'])
    assert_equal(response.payment_request.amount,
                 body.get('payment_request')['amount'])
    assert_equal(response.payment_request.app_fee,
                 body.get('payment_request')['app_fee'])
    assert_equal(response.payment_request.currency,
                 body.get('payment_request')['currency'])
    assert_equal(response.payment_request.description,
                 body.get('payment_request')['description'])
    assert_equal(response.payment_request.links,
                 body.get('payment_request')['links'])
    assert_equal(response.payment_request.metadata,
                 body.get('payment_request')['metadata'])
    assert_equal(response.payment_request.scheme,
                 body.get('payment_request')['scheme'])
    assert_equal(response.resources.customer,
                 body.get('resources')['customer'])
    assert_equal(response.resources.customer_bank_account,
                 body.get('resources')['customer_bank_account'])
    assert_equal(response.resources.customer_billing_detail,
                 body.get('resources')['customer_billing_detail'])

def test_timeout_billing_requests_confirm_payer_details_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['confirm_payer_details']
    with helpers.stub_timeout(fixture) as rsps:
      with assert_raises(requests.ConnectTimeout):
        response = helpers.client.billing_requests.confirm_payer_details(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))

def test_502_billing_requests_confirm_payer_details_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['confirm_payer_details']
    with helpers.stub_502(fixture) as rsps:
      with assert_raises(MalformedResponseError):
        response = helpers.client.billing_requests.confirm_payer_details(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))
  

@responses.activate
def test_billing_requests_cancel():
    fixture = helpers.load_fixture('billing_requests')['cancel']
    helpers.stub_response(fixture)
    response = helpers.client.billing_requests.cancel(*fixture['url_params'])
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.actions, body.get('actions'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.fallback_enabled, body.get('fallback_enabled'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.links.bank_authorisation,
                 body.get('links')['bank_authorisation'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.customer_billing_detail,
                 body.get('links')['customer_billing_detail'])
    assert_equal(response.links.mandate_request,
                 body.get('links')['mandate_request'])
    assert_equal(response.links.mandate_request_mandate,
                 body.get('links')['mandate_request_mandate'])
    assert_equal(response.links.payment_request,
                 body.get('links')['payment_request'])
    assert_equal(response.links.payment_request_payment,
                 body.get('links')['payment_request_payment'])
    assert_equal(response.mandate_request.currency,
                 body.get('mandate_request')['currency'])
    assert_equal(response.mandate_request.links,
                 body.get('mandate_request')['links'])
    assert_equal(response.mandate_request.metadata,
                 body.get('mandate_request')['metadata'])
    assert_equal(response.mandate_request.scheme,
                 body.get('mandate_request')['scheme'])
    assert_equal(response.mandate_request.verify,
                 body.get('mandate_request')['verify'])
    assert_equal(response.payment_request.amount,
                 body.get('payment_request')['amount'])
    assert_equal(response.payment_request.app_fee,
                 body.get('payment_request')['app_fee'])
    assert_equal(response.payment_request.currency,
                 body.get('payment_request')['currency'])
    assert_equal(response.payment_request.description,
                 body.get('payment_request')['description'])
    assert_equal(response.payment_request.links,
                 body.get('payment_request')['links'])
    assert_equal(response.payment_request.metadata,
                 body.get('payment_request')['metadata'])
    assert_equal(response.payment_request.scheme,
                 body.get('payment_request')['scheme'])
    assert_equal(response.resources.customer,
                 body.get('resources')['customer'])
    assert_equal(response.resources.customer_bank_account,
                 body.get('resources')['customer_bank_account'])
    assert_equal(response.resources.customer_billing_detail,
                 body.get('resources')['customer_billing_detail'])

def test_timeout_billing_requests_cancel_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['cancel']
    with helpers.stub_timeout(fixture) as rsps:
      with assert_raises(requests.ConnectTimeout):
        response = helpers.client.billing_requests.cancel(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))

def test_502_billing_requests_cancel_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['cancel']
    with helpers.stub_502(fixture) as rsps:
      with assert_raises(MalformedResponseError):
        response = helpers.client.billing_requests.cancel(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))
  

@responses.activate
def test_billing_requests_notify():
    fixture = helpers.load_fixture('billing_requests')['notify']
    helpers.stub_response(fixture)
    response = helpers.client.billing_requests.notify(*fixture['url_params'])
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.actions, body.get('actions'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.fallback_enabled, body.get('fallback_enabled'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.links.bank_authorisation,
                 body.get('links')['bank_authorisation'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.customer_billing_detail,
                 body.get('links')['customer_billing_detail'])
    assert_equal(response.links.mandate_request,
                 body.get('links')['mandate_request'])
    assert_equal(response.links.mandate_request_mandate,
                 body.get('links')['mandate_request_mandate'])
    assert_equal(response.links.payment_request,
                 body.get('links')['payment_request'])
    assert_equal(response.links.payment_request_payment,
                 body.get('links')['payment_request_payment'])
    assert_equal(response.mandate_request.currency,
                 body.get('mandate_request')['currency'])
    assert_equal(response.mandate_request.links,
                 body.get('mandate_request')['links'])
    assert_equal(response.mandate_request.metadata,
                 body.get('mandate_request')['metadata'])
    assert_equal(response.mandate_request.scheme,
                 body.get('mandate_request')['scheme'])
    assert_equal(response.mandate_request.verify,
                 body.get('mandate_request')['verify'])
    assert_equal(response.payment_request.amount,
                 body.get('payment_request')['amount'])
    assert_equal(response.payment_request.app_fee,
                 body.get('payment_request')['app_fee'])
    assert_equal(response.payment_request.currency,
                 body.get('payment_request')['currency'])
    assert_equal(response.payment_request.description,
                 body.get('payment_request')['description'])
    assert_equal(response.payment_request.links,
                 body.get('payment_request')['links'])
    assert_equal(response.payment_request.metadata,
                 body.get('payment_request')['metadata'])
    assert_equal(response.payment_request.scheme,
                 body.get('payment_request')['scheme'])
    assert_equal(response.resources.customer,
                 body.get('resources')['customer'])
    assert_equal(response.resources.customer_bank_account,
                 body.get('resources')['customer_bank_account'])
    assert_equal(response.resources.customer_billing_detail,
                 body.get('resources')['customer_billing_detail'])

def test_timeout_billing_requests_notify_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['notify']
    with helpers.stub_timeout(fixture) as rsps:
      with assert_raises(requests.ConnectTimeout):
        response = helpers.client.billing_requests.notify(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))

def test_502_billing_requests_notify_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['notify']
    with helpers.stub_502(fixture) as rsps:
      with assert_raises(MalformedResponseError):
        response = helpers.client.billing_requests.notify(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))
  

@responses.activate
def test_billing_requests_fallback():
    fixture = helpers.load_fixture('billing_requests')['fallback']
    helpers.stub_response(fixture)
    response = helpers.client.billing_requests.fallback(*fixture['url_params'])
    body = fixture['body']['billing_requests']

    assert_is_instance(response, resources.BillingRequest)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.actions, body.get('actions'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.fallback_enabled, body.get('fallback_enabled'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.status, body.get('status'))
    assert_equal(response.links.bank_authorisation,
                 body.get('links')['bank_authorisation'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.customer_billing_detail,
                 body.get('links')['customer_billing_detail'])
    assert_equal(response.links.mandate_request,
                 body.get('links')['mandate_request'])
    assert_equal(response.links.mandate_request_mandate,
                 body.get('links')['mandate_request_mandate'])
    assert_equal(response.links.payment_request,
                 body.get('links')['payment_request'])
    assert_equal(response.links.payment_request_payment,
                 body.get('links')['payment_request_payment'])
    assert_equal(response.mandate_request.currency,
                 body.get('mandate_request')['currency'])
    assert_equal(response.mandate_request.links,
                 body.get('mandate_request')['links'])
    assert_equal(response.mandate_request.metadata,
                 body.get('mandate_request')['metadata'])
    assert_equal(response.mandate_request.scheme,
                 body.get('mandate_request')['scheme'])
    assert_equal(response.mandate_request.verify,
                 body.get('mandate_request')['verify'])
    assert_equal(response.payment_request.amount,
                 body.get('payment_request')['amount'])
    assert_equal(response.payment_request.app_fee,
                 body.get('payment_request')['app_fee'])
    assert_equal(response.payment_request.currency,
                 body.get('payment_request')['currency'])
    assert_equal(response.payment_request.description,
                 body.get('payment_request')['description'])
    assert_equal(response.payment_request.links,
                 body.get('payment_request')['links'])
    assert_equal(response.payment_request.metadata,
                 body.get('payment_request')['metadata'])
    assert_equal(response.payment_request.scheme,
                 body.get('payment_request')['scheme'])
    assert_equal(response.resources.customer,
                 body.get('resources')['customer'])
    assert_equal(response.resources.customer_bank_account,
                 body.get('resources')['customer_bank_account'])
    assert_equal(response.resources.customer_billing_detail,
                 body.get('resources')['customer_billing_detail'])

def test_timeout_billing_requests_fallback_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['fallback']
    with helpers.stub_timeout(fixture) as rsps:
      with assert_raises(requests.ConnectTimeout):
        response = helpers.client.billing_requests.fallback(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))

def test_502_billing_requests_fallback_doesnt_retry():
    fixture = helpers.load_fixture('billing_requests')['fallback']
    with helpers.stub_502(fixture) as rsps:
      with assert_raises(MalformedResponseError):
        response = helpers.client.billing_requests.fallback(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

class Subscription(object):
    """A thin wrapper around a subscription, providing easy access to its
    attributes.

    Example:
      subscription = client.subscriptions.get()
      subscription.id
    """

    def __init__(self, attributes, api_response):
        self.attributes = attributes
        self.api_response = api_response

    @property
    def amount(self):
        return self.attributes.get('amount')
  

    @property
    def app_fee(self):
        return self.attributes.get('app_fee')
  

    @property
    def count(self):
        return self.attributes.get('count')
  

    @property
    def created_at(self):
        return self.attributes.get('created_at')
  

    @property
    def currency(self):
        return self.attributes.get('currency')
  

    @property
    def day_of_month(self):
        return self.attributes.get('day_of_month')
  

    @property
    def earliest_charge_date_after_resume(self):
        return self.attributes.get('earliest_charge_date_after_resume')
  

    @property
    def end_date(self):
        return self.attributes.get('end_date')
  

    @property
    def id(self):
        return self.attributes.get('id')
  

    @property
    def interval(self):
        return self.attributes.get('interval')
  

    @property
    def interval_unit(self):
        return self.attributes.get('interval_unit')
  

    @property
    def links(self):
        return self.Links(self.attributes.get('links'))
  

    @property
    def metadata(self):
        return self.attributes.get('metadata')
  

    @property
    def month(self):
        return self.attributes.get('month')
  

    @property
    def name(self):
        return self.attributes.get('name')
  

    @property
    def payment_reference(self):
        return self.attributes.get('payment_reference')
  

    @property
    def retry_if_possible(self):
        return self.attributes.get('retry_if_possible')
  

    @property
    def start_date(self):
        return self.attributes.get('start_date')
  

    @property
    def status(self):
        return self.attributes.get('status')
  

    @property
    def upcoming_payments(self):
        return self.attributes.get('upcoming_payments')
  


  

  

  

  

  

  

  

  

  

  

  

  
    class Links(object):
        """Wrapper for the response's 'links' attribute."""

        def __init__(self, attributes):
            self.attributes = attributes
    
        @property
        def mandate(self):
            return self.attributes.get('mandate')
    
  

  

  

  

  

  

  

  

  




# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

import json

import requests
import responses
from nose.tools import (
  assert_equal,
  assert_is_instance,
  assert_is_none,
  assert_is_not_none,
  assert_not_equal,
  assert_raises
)

from gocardless_pro.errors import MalformedResponseError
from gocardless_pro import resources
from gocardless_pro import list_response

from .. import helpers
  

@responses.activate
def test_redirect_flows_create():
    fixture = helpers.load_fixture('redirect_flows')['create']
    helpers.stub_response(fixture)
    response = helpers.client.redirect_flows.create(*fixture['url_params'])
    body = fixture['body']['redirect_flows']

    assert_is_instance(response, resources.RedirectFlow)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.confirmation_url, body.get('confirmation_url'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.description, body.get('description'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.mandate_reference, body.get('mandate_reference'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.redirect_url, body.get('redirect_url'))
    assert_equal(response.scheme, body.get('scheme'))
    assert_equal(response.session_token, body.get('session_token'))
    assert_equal(response.success_redirect_url, body.get('success_redirect_url'))
    assert_equal(response.links.billing_request,
                 body.get('links')['billing_request'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.mandate,
                 body.get('links')['mandate'])

@responses.activate
def test_redirect_flows_create_new_idempotency_key_for_each_call():
    fixture = helpers.load_fixture('redirect_flows')['create']
    helpers.stub_response(fixture)
    helpers.client.redirect_flows.create(*fixture['url_params'])
    helpers.client.redirect_flows.create(*fixture['url_params'])
    assert_not_equal(responses.calls[0].request.headers.get('Idempotency-Key'),
                     responses.calls[1].request.headers.get('Idempotency-Key'))

def test_timeout_redirect_flows_create_idempotency_conflict():
    create_fixture = helpers.load_fixture('redirect_flows')['create']
    get_fixture = helpers.load_fixture('redirect_flows')['get']
    with helpers.stub_timeout_then_idempotency_conflict(create_fixture, get_fixture) as rsps:
      response = helpers.client.redirect_flows.create(*create_fixture['url_params'])
      assert_equal(2, len(rsps.calls))

    assert_is_instance(response, resources.RedirectFlow)

@responses.activate
def test_timeout_redirect_flows_create_retries():
    fixture = helpers.load_fixture('redirect_flows')['create']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.redirect_flows.create(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['redirect_flows']

    assert_is_instance(response, resources.RedirectFlow)

def test_502_redirect_flows_create_retries():
    fixture = helpers.load_fixture('redirect_flows')['create']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.redirect_flows.create(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['redirect_flows']

    assert_is_instance(response, resources.RedirectFlow)
  

@responses.activate
def test_redirect_flows_get():
    fixture = helpers.load_fixture('redirect_flows')['get']
    helpers.stub_response(fixture)
    response = helpers.client.redirect_flows.get(*fixture['url_params'])
    body = fixture['body']['redirect_flows']

    assert_is_instance(response, resources.RedirectFlow)
    assert_is_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.confirmation_url, body.get('confirmation_url'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.description, body.get('description'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.mandate_reference, body.get('mandate_reference'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.redirect_url, body.get('redirect_url'))
    assert_equal(response.scheme, body.get('scheme'))
    assert_equal(response.session_token, body.get('session_token'))
    assert_equal(response.success_redirect_url, body.get('success_redirect_url'))
    assert_equal(response.links.billing_request,
                 body.get('links')['billing_request'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.mandate,
                 body.get('links')['mandate'])

@responses.activate
def test_timeout_redirect_flows_get_retries():
    fixture = helpers.load_fixture('redirect_flows')['get']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.redirect_flows.get(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['redirect_flows']

    assert_is_instance(response, resources.RedirectFlow)

def test_502_redirect_flows_get_retries():
    fixture = helpers.load_fixture('redirect_flows')['get']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.redirect_flows.get(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['redirect_flows']

    assert_is_instance(response, resources.RedirectFlow)
  

@responses.activate
def test_redirect_flows_complete():
    fixture = helpers.load_fixture('redirect_flows')['complete']
    helpers.stub_response(fixture)
    response = helpers.client.redirect_flows.complete(*fixture['url_params'])
    body = fixture['body']['redirect_flows']

    assert_is_instance(response, resources.RedirectFlow)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.confirmation_url, body.get('confirmation_url'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.description, body.get('description'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.mandate_reference, body.get('mandate_reference'))
    assert_equal(response.metadata, body.get('metadata'))
    assert_equal(response.redirect_url, body.get('redirect_url'))
    assert_equal(response.scheme, body.get('scheme'))
    assert_equal(response.session_token, body.get('session_token'))
    assert_equal(response.success_redirect_url, body.get('success_redirect_url'))
    assert_equal(response.links.billing_request,
                 body.get('links')['billing_request'])
    assert_equal(response.links.creditor,
                 body.get('links')['creditor'])
    assert_equal(response.links.customer,
                 body.get('links')['customer'])
    assert_equal(response.links.customer_bank_account,
                 body.get('links')['customer_bank_account'])
    assert_equal(response.links.mandate,
                 body.get('links')['mandate'])

def test_timeout_redirect_flows_complete_doesnt_retry():
    fixture = helpers.load_fixture('redirect_flows')['complete']
    with helpers.stub_timeout(fixture) as rsps:
      with assert_raises(requests.ConnectTimeout):
        response = helpers.client.redirect_flows.complete(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))

def test_502_redirect_flows_complete_doesnt_retry():
    fixture = helpers.load_fixture('redirect_flows')['complete']
    with helpers.stub_502(fixture) as rsps:
      with assert_raises(MalformedResponseError):
        response = helpers.client.redirect_flows.complete(*fixture['url_params'])
      assert_equal(1, len(rsps.calls))
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

class Institution(object):
    """A thin wrapper around a institution, providing easy access to its
    attributes.

    Example:
      institution = client.institutions.get()
      institution.id
    """

    def __init__(self, attributes, api_response):
        self.attributes = attributes
        self.api_response = api_response

    @property
    def country_code(self):
        return self.attributes.get('country_code')
  

    @property
    def icon_url(self):
        return self.attributes.get('icon_url')
  

    @property
    def id(self):
        return self.attributes.get('id')
  

    @property
    def logo_url(self):
        return self.attributes.get('logo_url')
  

    @property
    def name(self):
        return self.attributes.get('name')
  


  

  

  

  

  




# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

class Customer(object):
    """A thin wrapper around a customer, providing easy access to its
    attributes.

    Example:
      customer = client.customers.get()
      customer.id
    """

    def __init__(self, attributes, api_response):
        self.attributes = attributes
        self.api_response = api_response

    @property
    def address_line1(self):
        return self.attributes.get('address_line1')
  

    @property
    def address_line2(self):
        return self.attributes.get('address_line2')
  

    @property
    def address_line3(self):
        return self.attributes.get('address_line3')
  

    @property
    def city(self):
        return self.attributes.get('city')
  

    @property
    def company_name(self):
        return self.attributes.get('company_name')
  

    @property
    def country_code(self):
        return self.attributes.get('country_code')
  

    @property
    def created_at(self):
        return self.attributes.get('created_at')
  

    @property
    def danish_identity_number(self):
        return self.attributes.get('danish_identity_number')
  

    @property
    def email(self):
        return self.attributes.get('email')
  

    @property
    def family_name(self):
        return self.attributes.get('family_name')
  

    @property
    def given_name(self):
        return self.attributes.get('given_name')
  

    @property
    def id(self):
        return self.attributes.get('id')
  

    @property
    def language(self):
        return self.attributes.get('language')
  

    @property
    def metadata(self):
        return self.attributes.get('metadata')
  

    @property
    def phone_number(self):
        return self.attributes.get('phone_number')
  

    @property
    def postal_code(self):
        return self.attributes.get('postal_code')
  

    @property
    def region(self):
        return self.attributes.get('region')
  

    @property
    def swedish_identity_number(self):
        return self.attributes.get('swedish_identity_number')
  


  

  

  

  

  

  

  

  

  

  

  

  

  

  

  

  

  

  




# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class MandatePdfsService(base_service.BaseService):
    """Service class that provides access to the mandate_pdfs
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.MandatePdf
    RESOURCE_NAME = 'mandate_pdfs'


    def create(self,params=None, headers=None):
        """Create a mandate PDF.

        Generates a PDF mandate and returns its temporary URL.
        
        Customer and bank account details can be left blank (for a blank
        mandate), provided manually, or inferred from the ID of an existing
        [mandate](#core-endpoints-mandates).
        
        By default, we'll generate PDF mandates in English.
        
        To generate a PDF mandate in another language, set the
        `Accept-Language` header when creating the PDF mandate to the relevant
        [ISO 639-1](http://en.wikipedia.org/wiki/List_of_ISO_639-1_codes)
        language code supported for the scheme.
        
        | Scheme           | Supported languages                               
                                                                               
                          |
        | :--------------- |
        :-------------------------------------------------------------------------------------------------------------------------------------------
        |
        | ACH              | English (`en`)                                    
                                                                               
                          |
        | Autogiro         | English (`en`), Swedish (`sv`)                    
                                                                               
                          |
        | Bacs             | English (`en`)                                    
                                                                               
                          |
        | BECS             | English (`en`)                                    
                                                                               
                          |
        | BECS NZ          | English (`en`)                                    
                                                                               
                          |
        | Betalingsservice | Danish (`da`), English (`en`)                     
                                                                               
                          |
        | PAD              | English (`en`)                                    
                                                                               
                          |
        | SEPA Core        | Danish (`da`), Dutch (`nl`), English (`en`),
        French (`fr`), German (`de`), Italian (`it`), Portuguese (`pt`),
        Spanish (`es`), Swedish (`sv`) |

        Args:
              params (dict, optional): Request body.

        Returns:
              MandatePdf
        """
        path = '/mandate_pdfs'
        
        if params is not None:
            params = {self._envelope_key(): params}

        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class BlocksService(base_service.BaseService):
    """Service class that provides access to the blocks
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.Block
    RESOURCE_NAME = 'blocks'


    def create(self,params=None, headers=None):
        """Create a block.

        Creates a new Block of a given type. By default it will be active.

        Args:
              params (dict, optional): Request body.

        Returns:
              Block
        """
        path = '/blocks'
        
        if params is not None:
            params = {self._envelope_key(): params}

        try:
          response = self._perform_request('POST', path, params, headers,
                                            retry_failures=True)
        except errors.IdempotentCreationConflictError as err:
          if self.raise_on_idempotency_conflict:
            raise err
          return self.get(identity=err.conflicting_resource_id,
                          params=params,
                          headers=headers)
        return self._resource_for(response)
  

    def get(self,identity,params=None, headers=None):
        """Get a single block.

        Retrieves the details of an existing block.

        Args:
              identity (string): Unique identifier, beginning with "BLC".
              params (dict, optional): Query string parameters.

        Returns:
              Block
        """
        path = self._sub_url_params('/blocks/:identity', {
          
            'identity': identity,
          })
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def list(self,params=None, headers=None):
        """List multiple blocks.

        Returns a [cursor-paginated](#api-usage-cursor-pagination) list of your
        blocks.

        Args:
              params (dict, optional): Query string parameters.

        Returns:
              ListResponse of Block instances
        """
        path = '/blocks'
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)

    def all(self, params=None):
        if params is None:
            params = {}
        return Paginator(self, params)
    
  

    def disable(self,identity,params=None, headers=None):
        """Disable a block.

        Disables a block so that it no longer will prevent mandate creation.

        Args:
              identity (string): Unique identifier, beginning with "BLC".
              params (dict, optional): Request body.

        Returns:
              Block
        """
        path = self._sub_url_params('/blocks/:identity/actions/disable', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def enable(self,identity,params=None, headers=None):
        """Enable a block.

        Enables a previously disabled block so that it will prevent mandate
        creation

        Args:
              identity (string): Unique identifier, beginning with "BLC".
              params (dict, optional): Request body.

        Returns:
              Block
        """
        path = self._sub_url_params('/blocks/:identity/actions/enable', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def block_by_ref(self,params=None, headers=None):
        """Create blocks by reference.

        Creates new blocks for a given reference. By default blocks will be
        active.
        Returns 201 if at least one block was created. Returns 200 if there
        were no new
        blocks created.

        Args:
              params (dict, optional): Request body.

        Returns:
              ListResponse of Block instances
        """
        path = '/block_by_ref'
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class BillingRequestFlowsService(base_service.BaseService):
    """Service class that provides access to the billing_request_flows
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.BillingRequestFlow
    RESOURCE_NAME = 'billing_request_flows'


    def create(self,params=None, headers=None):
        """Create a Billing Request Flow.

        Creates a new billing request flow.

        Args:
              params (dict, optional): Request body.

        Returns:
              BillingRequestFlow
        """
        path = '/billing_request_flows'
        
        if params is not None:
            params = {self._envelope_key(): params}

        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def initialise(self,identity,params=None, headers=None):
        """Initialise a Billing Request Flow.

        Returns the flow having generated a fresh session token which can be
        used to power
        integrations that manipulate the flow.

        Args:
              identity (string): Unique identifier, beginning with "BRF".
              params (dict, optional): Request body.

        Returns:
              BillingRequestFlow
        """
        path = self._sub_url_params('/billing_request_flows/:identity/actions/initialise', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

import json

import requests
import responses
from nose.tools import (
  assert_equal,
  assert_is_instance,
  assert_is_none,
  assert_is_not_none,
  assert_not_equal,
  assert_raises
)

from gocardless_pro.errors import MalformedResponseError
from gocardless_pro import resources
from gocardless_pro import list_response

from .. import helpers
  

@responses.activate
def test_bank_authorisations_get():
    fixture = helpers.load_fixture('bank_authorisations')['get']
    helpers.stub_response(fixture)
    response = helpers.client.bank_authorisations.get(*fixture['url_params'])
    body = fixture['body']['bank_authorisations']

    assert_is_instance(response, resources.BankAuthorisation)
    assert_is_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.authorisation_type, body.get('authorisation_type'))
    assert_equal(response.authorised_at, body.get('authorised_at'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.expires_at, body.get('expires_at'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.last_visited_at, body.get('last_visited_at'))
    assert_equal(response.redirect_uri, body.get('redirect_uri'))
    assert_equal(response.url, body.get('url'))
    assert_equal(response.links.billing_request,
                 body.get('links')['billing_request'])
    assert_equal(response.links.institution,
                 body.get('links')['institution'])

@responses.activate
def test_timeout_bank_authorisations_get_retries():
    fixture = helpers.load_fixture('bank_authorisations')['get']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.bank_authorisations.get(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['bank_authorisations']

    assert_is_instance(response, resources.BankAuthorisation)

def test_502_bank_authorisations_get_retries():
    fixture = helpers.load_fixture('bank_authorisations')['get']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.bank_authorisations.get(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['bank_authorisations']

    assert_is_instance(response, resources.BankAuthorisation)
  

@responses.activate
def test_bank_authorisations_create():
    fixture = helpers.load_fixture('bank_authorisations')['create']
    helpers.stub_response(fixture)
    response = helpers.client.bank_authorisations.create(*fixture['url_params'])
    body = fixture['body']['bank_authorisations']

    assert_is_instance(response, resources.BankAuthorisation)
    assert_is_not_none(responses.calls[-1].request.headers.get('Idempotency-Key'))
    assert_equal(response.authorisation_type, body.get('authorisation_type'))
    assert_equal(response.authorised_at, body.get('authorised_at'))
    assert_equal(response.created_at, body.get('created_at'))
    assert_equal(response.expires_at, body.get('expires_at'))
    assert_equal(response.id, body.get('id'))
    assert_equal(response.last_visited_at, body.get('last_visited_at'))
    assert_equal(response.redirect_uri, body.get('redirect_uri'))
    assert_equal(response.url, body.get('url'))
    assert_equal(response.links.billing_request,
                 body.get('links')['billing_request'])
    assert_equal(response.links.institution,
                 body.get('links')['institution'])

@responses.activate
def test_bank_authorisations_create_new_idempotency_key_for_each_call():
    fixture = helpers.load_fixture('bank_authorisations')['create']
    helpers.stub_response(fixture)
    helpers.client.bank_authorisations.create(*fixture['url_params'])
    helpers.client.bank_authorisations.create(*fixture['url_params'])
    assert_not_equal(responses.calls[0].request.headers.get('Idempotency-Key'),
                     responses.calls[1].request.headers.get('Idempotency-Key'))

def test_timeout_bank_authorisations_create_idempotency_conflict():
    create_fixture = helpers.load_fixture('bank_authorisations')['create']
    get_fixture = helpers.load_fixture('bank_authorisations')['get']
    with helpers.stub_timeout_then_idempotency_conflict(create_fixture, get_fixture) as rsps:
      response = helpers.client.bank_authorisations.create(*create_fixture['url_params'])
      assert_equal(2, len(rsps.calls))

    assert_is_instance(response, resources.BankAuthorisation)

@responses.activate
def test_timeout_bank_authorisations_create_retries():
    fixture = helpers.load_fixture('bank_authorisations')['create']
    with helpers.stub_timeout_then_response(fixture) as rsps:
      response = helpers.client.bank_authorisations.create(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['bank_authorisations']

    assert_is_instance(response, resources.BankAuthorisation)

def test_502_bank_authorisations_create_retries():
    fixture = helpers.load_fixture('bank_authorisations')['create']
    with helpers.stub_502_then_response(fixture) as rsps:
      response = helpers.client.bank_authorisations.create(*fixture['url_params'])
      assert_equal(2, len(rsps.calls))
      assert_equal(rsps.calls[0].request.headers.get('Idempotency-Key'),
                   rsps.calls[1].request.headers.get('Idempotency-Key'))
    body = fixture['body']['bank_authorisations']

    assert_is_instance(response, resources.BankAuthorisation)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class PayerAuthorisationsService(base_service.BaseService):
    """Service class that provides access to the payer_authorisations
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.PayerAuthorisation
    RESOURCE_NAME = 'payer_authorisations'


    def get(self,identity,params=None, headers=None):
        """Get a single Payer Authorisation.

        Retrieves the details of a single existing Payer Authorisation. It can
        be used for polling the status of a Payer Authorisation.

        Args:
              identity (string): Unique identifier, beginning with "PA".
              params (dict, optional): Query string parameters.

        Returns:
              PayerAuthorisation
        """
        path = self._sub_url_params('/payer_authorisations/:identity', {
          
            'identity': identity,
          })
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def create(self,params=None, headers=None):
        """Create a Payer Authorisation.

        Creates a Payer Authorisation. The resource is saved to the database
        even if incomplete. An empty array of incomplete_fields means that the
        resource is valid. The ID of the resource is used for the other
        actions. This endpoint has been designed this way so you do not need to
        save any payer data on your servers or the browser while still being
        able to implement a progressive solution, such as a multi-step form.

        Args:
              params (dict, optional): Request body.

        Returns:
              PayerAuthorisation
        """
        path = '/payer_authorisations'
        
        if params is not None:
            params = {self._envelope_key(): params}

        try:
          response = self._perform_request('POST', path, params, headers,
                                            retry_failures=True)
        except errors.IdempotentCreationConflictError as err:
          if self.raise_on_idempotency_conflict:
            raise err
          return self.get(identity=err.conflicting_resource_id,
                          params=params,
                          headers=headers)
        return self._resource_for(response)
  

    def update(self,identity,params=None, headers=None):
        """Update a Payer Authorisation.

        Updates a Payer Authorisation. Updates the Payer Authorisation with the
        request data. Can be invoked as many times as needed. Only fields
        present in the request will be modified. An empty array of
        incomplete_fields means that the resource is valid. This endpoint has
        been designed this way so you do not need to save any payer data on
        your servers or the browser while still being able to implement a
        progressive solution, such a multi-step form. <p class="notice"> Note
        that in order to update the `metadata` attribute values it must be sent
        completely as it overrides the previously existing values. </p>

        Args:
              identity (string): Unique identifier, beginning with "PA".
              params (dict, optional): Request body.

        Returns:
              PayerAuthorisation
        """
        path = self._sub_url_params('/payer_authorisations/:identity', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {self._envelope_key(): params}

        response = self._perform_request('PUT', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def submit(self,identity,params=None, headers=None):
        """Submit a Payer Authorisation.

        Submits all the data previously pushed to this PayerAuthorisation for
        verification. This time, a 200 HTTP status is returned if the resource
        is valid and a 422 error response in case of validation errors. After
        it is successfully submitted, the Payer Authorisation can no longer be
        edited.

        Args:
              identity (string): Unique identifier, beginning with "PA".
              params (dict, optional): Request body.

        Returns:
              PayerAuthorisation
        """
        path = self._sub_url_params('/payer_authorisations/:identity/actions/submit', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  

    def confirm(self,identity,params=None, headers=None):
        """Confirm a Payer Authorisation.

        Confirms the Payer Authorisation, indicating that the resources are
        ready to be created.
        A Payer Authorisation cannot be confirmed if it hasn't been submitted
        yet.
        
        <p class="notice">
          The main use of the confirm endpoint is to enable integrators to
        acknowledge the end of the setup process. 
          They might want to make the payers go through some other steps after
        they go through our flow or make them go through the necessary
        verification mechanism (upcoming feature).
        </p>

        Args:
              identity (string): Unique identifier, beginning with "PA".
              params (dict, optional): Request body.

        Returns:
              PayerAuthorisation
        """
        path = self._sub_url_params('/payer_authorisations/:identity/actions/confirm', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class CustomersService(base_service.BaseService):
    """Service class that provides access to the customers
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.Customer
    RESOURCE_NAME = 'customers'


    def create(self,params=None, headers=None):
        """Create a customer.

        Creates a new customer object.

        Args:
              params (dict, optional): Request body.

        Returns:
              Customer
        """
        path = '/customers'
        
        if params is not None:
            params = {self._envelope_key(): params}

        try:
          response = self._perform_request('POST', path, params, headers,
                                            retry_failures=True)
        except errors.IdempotentCreationConflictError as err:
          if self.raise_on_idempotency_conflict:
            raise err
          return self.get(identity=err.conflicting_resource_id,
                          params=params,
                          headers=headers)
        return self._resource_for(response)
  

    def list(self,params=None, headers=None):
        """List customers.

        Returns a [cursor-paginated](#api-usage-cursor-pagination) list of your
        customers.

        Args:
              params (dict, optional): Query string parameters.

        Returns:
              ListResponse of Customer instances
        """
        path = '/customers'
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)

    def all(self, params=None):
        if params is None:
            params = {}
        return Paginator(self, params)
    
  

    def get(self,identity,params=None, headers=None):
        """Get a single customer.

        Retrieves the details of an existing customer.

        Args:
              identity (string): Unique identifier, beginning with "CU".
              params (dict, optional): Query string parameters.

        Returns:
              Customer
        """
        path = self._sub_url_params('/customers/:identity', {
          
            'identity': identity,
          })
        

        response = self._perform_request('GET', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def update(self,identity,params=None, headers=None):
        """Update a customer.

        Updates a customer object. Supports all of the fields supported when
        creating a customer.

        Args:
              identity (string): Unique identifier, beginning with "CU".
              params (dict, optional): Request body.

        Returns:
              Customer
        """
        path = self._sub_url_params('/customers/:identity', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {self._envelope_key(): params}

        response = self._perform_request('PUT', path, params, headers,
                                         retry_failures=True)
        return self._resource_for(response)
  

    def remove(self,identity,params=None, headers=None):
        """Remove a customer.

        Removed customers will not appear in search results or lists of
        customers (in our API
        or exports), and it will not be possible to load an individually
        removed customer by
        ID.
        
        <p class="restricted-notice"><strong>The action of removing a customer
        cannot be reversed, so please use with care.</strong></p>

        Args:
              identity (string): Unique identifier, beginning with "CU".
              params (dict, optional): Request body.

        Returns:
              Customer
        """
        path = self._sub_url_params('/customers/:identity', {
          
            'identity': identity,
          })
        
        response = self._perform_request('DELETE', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  



# WARNING: Do not edit by hand, this file was generated by Crank:
#
#   https://github.com/gocardless/crank
#

from . import base_service
from .. import resources
from ..paginator import Paginator
from .. import errors

class ScenarioSimulatorsService(base_service.BaseService):
    """Service class that provides access to the scenario_simulators
    endpoints of the GoCardless Pro API.
    """

    RESOURCE_CLASS = resources.ScenarioSimulator
    RESOURCE_NAME = 'scenario_simulators'


    def run(self,identity,params=None, headers=None):
        """Simulate a scenario.

        Runs the specific scenario simulator against the specific resource

        Args:
              identity (string): The unique identifier of the simulator, used to initiate simulations. One of:
<ul>
<li>`creditor_verification_status_action_required`: Sets a creditor's `verification status` to `action required`, meaning that the creditor must provide further information to GoCardless in order to verify their account to receive payouts.</li>
<li>`creditor_verification_status_in_review`: Sets a creditor's `verification status` to `in review`, meaning that the creditor has provided all of the information requested by GoCardless to verify their account, and is now awaiting review.</li>
<li>`creditor_verification_status_successful`: Sets a creditor's `verification status` to `successful`, meaning that the creditor is fully verified and can receive payouts.</li>
<li>`payment_confirmed`: Transitions a payment through to `confirmed`. It must start in the `pending_submission` state, and its mandate must be in the `activated` state (unless it is a payment for ACH, BECS, BECS_NZ or SEPA, in which cases the mandate may be `pending_submission`, since their mandates are submitted with their first payment).</li>
<li>`payment_paid_out`: Transitions a payment through to `paid_out`, having been collected successfully and paid out to you. It must start in the `pending_submission` state, and its mandate must be in the `activated` state (unless it is a payment for ACH, BECS, BECS_NZ or SEPA, in which cases the mandate may be `pending_submission`, since their mandates are submitted with their first payment).</li>
<li>`payment_failed`: Transitions a payment through to `failed`. It must start in the `pending_submission` state, and its mandate must be in the `activated` state (unless it is a payment for ACH, BECS, BECS_NZ or SEPA, in which cases the mandate may be `pending_submission`, since their mandates are submitted with their first payment).</li>
<li>`payment_charged_back`: Behaves the same as the `payout_paid_out` simulator, except that the payment is transitioned to `charged_back` after it is paid out, having been charged back by the customer.</li>
<li>`payment_chargeback_settled`: Behaves the same as the `payment_charged_back` simulator, except that the charged back payment is additionally included as a debit item in a payout, thereby settling the charged back payment.</li>
<li>`payment_late_failure`: Transitions a payment through to `late_failure`, having been apparently collected successfully beforehand. It must start in the `pending_submission` state, and its mandate must be in the `activated` state (unless it is a payment for ACH, BECS, BECS_NZ or SEPA, in which cases the mandate may be `pending_submission`, since their mandates are submitted with their first payment). Not compatible with Autogiro mandates.</li>
<li>`payment_late_failure_settled`: Behaves the same as the `payment_late_failure` simulator, except that the late failure is additionally included as a debit item in a payout, thereby settling the late failure.</li>
<li>`payment_submitted`: Transitions a payment to `submitted`, without proceeding any further. It must start in the `pending_submission` state.</li>
<li>`mandate_activated`: Transitions a mandate through to `activated`, having been submitted to the banks and set up successfully. It must start in the `pending_submission` state. Not compatible with ACH, BECS, BECS_NZ and SEPA mandates, which are submitted and activated with their first payment.</li>
<li>`mandate_customer_approval_granted`: Transitions a mandate through to `pending_submission`, as if the customer approved the mandate creation. It must start in the `pending_customer_approval` state. Compatible only with Bacs and SEPA mandates, which support customer signatures on the mandate. All payments associated with the mandate will be transitioned to `pending_submission`. All subscriptions associated with the mandate will become `active`.</li>
<li>`mandate_customer_approval_skipped`: Transitions a mandate through to `pending_submission`, as if the customer skipped the mandate approval during the mandate creation process. It must start in the `pending_customer_approval` state. Compatible only with Bacs and SEPA mandates, which support customer signatures on the mandate. All payments associated with the mandate will be transitioned to `pending_submission`. All subscriptions associated with the mandate will become `active`.</li>
<li>`mandate_failed`: Transitions a mandate through to `failed`, having been submitted to the banks but found to be invalid (for example due to invalid bank details). It must start in the `pending_submission` or `submitted` states. Not compatible with SEPA mandates, which are submitted with their first payment.</li>
<li>`mandate_expired`: Transitions a mandate through to `expired`, having been submitted to the banks, set up successfully and then expired because no collection attempts were made against it for longer than the scheme's dormancy period (13 months for Bacs, 3 years for SEPA, 15 months for ACH, Betalingsservice, and BECS). It must start in the `pending_submission` state. Not compatible with Autogiro, BECS NZ, and PAD mandates, which do not expire.</li>
<li>`mandate_transferred`: Transitions a mandate through to `transferred`, having been submitted to the banks, set up successfully and then moved to a new bank account due to the customer using the UK's Current Account Switching Service (CASS). It must start in the `pending_submission` state. Only compatible with Bacs mandates.</li>
<li>`mandate_transferred_with_resubmission`: Transitions a mandate through `transferred` and resubmits it to the banks, can be caused be the UK's Current Account Switching Service (CASS) or when a customer contacts GoCardless to change their bank details. It must start in the `pending_submission` state. Only compatible with Bacs, SEPA and Autogiro mandates.</li>
<li>`refund_paid`: Transitions a refund to `paid`. It must start in either the `pending_submission` or `submitted` state.</li>
<li>`refund_settled`: Transitions a refund to `paid`, if it's not already, then generates a payout that includes the refund, thereby settling the funds. It must start in one of `pending_submission`, `submitted` or `paid` states.</li>
<li>`refund_bounced`: Transitions a refund to `bounced`. It must start in either the `pending_submission`, `submitted`, or `paid` state.</li>
<li>`refund_returned`: Transitions a refund to `refund_returned`. The refund must start in `pending_submission`.</li>
<li>`payout_bounced`: Transitions a payout to `bounced`. It must start in the `paid` state.</li>
<li>`billing_request_fulfilled`: Authorises the billing request, fulfils it, and moves the associated payment to `failed`. The billing request must be in the `pending` state, with all actions completed except for `bank_authorisation`. Only billing requests with a `payment_request` are supported.</li>
<li>`billing_request_fulfilled_and_payment_failed`: Authorises the billing request, fulfils it, and moves the associated payment to `failed`. The billing request must be in the `pending` state, with all actions completed except for `bank_authorisation`. Only billing requests with a `payment_request` are supported.</li>
<li>`billing_request_fulfilled_and_payment_paid_out`: Authorises the billing request, fulfils it, and moves the associated payment to `paid_out`. The billing request must be in the `pending` state, with all actions completed except for `bank_authorisation`. Only billing requests with a `payment_request` are supported.</li>
</ul>
              params (dict, optional): Request body.

        Returns:
              ScenarioSimulator
        """
        path = self._sub_url_params('/scenario_simulators/:identity/actions/run', {
          
            'identity': identity,
          })
        
        if params is not None:
            params = {'data': params}
        response = self._perform_request('POST', path, params, headers,
                                         retry_failures=False)
        return self._resource_for(response)
  
